import datetime
import os
import uuid
from ctypes import (
    POINTER,
    Structure,
    Union,
    byref,
    c_byte,
    c_char,
    c_char_p,
    c_int,
    c_long,
    c_longlong,
    c_short,
    c_uint,
    c_ulong,
    c_ulonglong,
    c_ushort,
    c_void_p,
    cdll,
    create_string_buffer,
    sizeof,
)

from .utils import unix_epoch

libc = cdll.LoadLibrary("/usr/lib/libc.dylib")

# Constants
FSOPT_NOFOLLOW = 0x00000001
FSOPT_NOINMEMUPDATE = 0x00000002
FSOPT_REPORT_FULLSIZE = 0x00000004
FSOPT_PACK_INVAL_ATTRS = 0x00000008
FSOPT_ATTR_CMN_EXTENDED = 0x00000020
FSOPT_RETURN_REALDEV = 0x00000200

VOL_CAPABILITIES_FORMAT = 0
VOL_CAPABILITIES_INTERFACES = 1

VOL_CAP_FMT_PERSISTENTOBJECTIDS = 0x00000001
VOL_CAP_FMT_SYMBOLICLINKS = 0x00000002
VOL_CAP_FMT_HARDLINKS = 0x00000004
VOL_CAP_FMT_JOURNAL = 0x00000008
VOL_CAP_FMT_JOURNAL_ACTIVE = 0x00000010
VOL_CAP_FMT_NO_ROOT_TIMES = 0x00000020
VOL_CAP_FMT_SPARSE_FILES = 0x00000040
VOL_CAP_FMT_ZERO_RUNS = 0x00000080
VOL_CAP_FMT_CASE_SENSITIVE = 0x00000100
VOL_CAP_FMT_CASE_PRESERVING = 0x00000200
VOL_CAP_FMT_FAST_STATFS = 0x00000400
VOL_CAP_FMT_2TB_FILESIZE = 0x00000800
VOL_CAP_FMT_OPENDENYMODES = 0x00001000
VOL_CAP_FMT_HIDDEN_FILES = 0x00002000
VOL_CAP_FMT_PATH_FROM_ID = 0x00004000
VOL_CAP_FMT_NO_VOLUME_SIZES = 0x00008000
VOL_CAP_FMT_DECMPFS_COMPRESSION = 0x00010000
VOL_CAP_FMT_64BIT_OBJECT_IDS = 0x00020000

VOL_CAP_INT_SEARCHFS = 0x00000001
VOL_CAP_INT_ATTRLIST = 0x00000002
VOL_CAP_INT_NFSEXPORT = 0x00000004
VOL_CAP_INT_READDIRATTR = 0x00000008
VOL_CAP_INT_EXCHANGEDATA = 0x00000010
VOL_CAP_INT_COPYFILE = 0x00000020
VOL_CAP_INT_ALLOCATE = 0x00000040
VOL_CAP_INT_VOL_RENAME = 0x00000080
VOL_CAP_INT_ADVLOCK = 0x00000100
VOL_CAP_INT_FLOCK = 0x00000200
VOL_CAP_INT_EXTENDED_SECURITY = 0x00000400
VOL_CAP_INT_USERACCESS = 0x00000800
VOL_CAP_INT_MANLOCK = 0x00001000
VOL_CAP_INT_NAMEDSTREAMS = 0x00002000
VOL_CAP_INT_EXTENDED_ATTR = 0x00004000
VOL_CAP_INT_CLONE = 0x00010000
VOL_CAP_INT_SNAPSHOT = 0x00020000
VOL_CAP_INT_RENAME_SWAP = 0x00040000
VOL_CAP_INT_RENAME_EXCL = 0x00080000
VOL_CAP_INT_RENAME_OPENFAIL = 0x00100000

ATTR_CMN_NAME = 0x00000001
ATTR_CMN_DEVID = 0x00000002
ATTR_CMN_FSID = 0x00000004
ATTR_CMN_OBJTYPE = 0x00000008
ATTR_CMN_OBJTAG = 0x00000010
ATTR_CMN_OBJID = 0x00000020
ATTR_CMN_OBJPERMANENTID = 0x00000040
ATTR_CMN_PAROBJID = 0x00000080
ATTR_CMN_SCRIPT = 0x00000100
ATTR_CMN_CRTIME = 0x00000200
ATTR_CMN_MODTIME = 0x00000400
ATTR_CMN_CHGTIME = 0x00000800
ATTR_CMN_ACCTIME = 0x00001000
ATTR_CMN_BKUPTIME = 0x00002000
ATTR_CMN_FNDRINFO = 0x00004000
ATTR_CMN_OWNERID = 0x00008000
ATTR_CMN_GRPID = 0x00010000
ATTR_CMN_ACCESSMASK = 0x00020000
ATTR_CMN_FLAGS = 0x00040000
ATTR_CMN_GEN_COUNT = 0x00080000
ATTR_CMN_DOCUMENT_ID = 0x00100000
ATTR_CMN_USERACCESS = 0x00200000
ATTR_CMN_EXTENDED_SECURITY = 0x00400000
ATTR_CMN_UUID = 0x00800000
ATTR_CMN_GRPUUID = 0x01000000
ATTR_CMN_FILEID = 0x02000000
ATTR_CMN_PARENTID = 0x04000000
ATTR_CMN_FULLPATH = 0x08000000
ATTR_CMN_ADDEDTIME = 0x10000000
ATTR_CMN_ERROR = 0x20000000
ATTR_CMN_DATA_PROTECT_FLAGS = 0x40000000
ATTR_CMN_RETURNED_ATTRS = 0x80000000
ATTR_CMN_ALL_ATTRS = 0xFFFFFFFF

ATTR_CMN_VALIDMASK = 0xFFFFFFFF
ATTR_CMN_SETMASK = 0x51C7FF00
ATTR_CMN_VOLSETMASK = 0x00006700

ATTR_VOL_FSTYPE = 0x00000001
ATTR_VOL_SIGNATURE = 0x00000002
ATTR_VOL_SIZE = 0x00000004
ATTR_VOL_SPACEFREE = 0x00000008
ATTR_VOL_SPACEAVAIL = 0x00000010
ATTR_VOL_MINALLOCATION = 0x00000020
ATTR_VOL_ALLOCATIONCLUMP = 0x00000040
ATTR_VOL_IOBLOCKSIZE = 0x00000080
ATTR_VOL_OBJCOUNT = 0x00000100
ATTR_VOL_FILECOUNT = 0x00000200
ATTR_VOL_DIRCOUNT = 0x00000400
ATTR_VOL_MAXOBJCOUNT = 0x00000800
ATTR_VOL_MOUNTPOINT = 0x00001000
ATTR_VOL_NAME = 0x00002000
ATTR_VOL_MOUNTFLAGS = 0x00004000
ATTR_VOL_MOUNTEDDEVICE = 0x00008000
ATTR_VOL_ENCODINGSUSED = 0x00010000
ATTR_VOL_CAPABILITIES = 0x00020000
ATTR_VOL_UUID = 0x00040000
ATTR_VOL_QUOTA_SIZE = 0x10000000
ATTR_VOL_RESERVED_SIZE = 0x20000000
ATTR_VOL_ATTRIBUTES = 0x40000000
ATTR_VOL_INFO = 0x80000000
ATTR_VOL_ALL_ATTRS = 0xF007FFFF

ATTR_DIR_LINKCOUNT = 0x00000001
ATTR_DIR_ENTRYCOUNT = 0x00000002
ATTR_DIR_MOUNTSTATUS = 0x00000004
DIR_MNTSTATUS_MNTPOINT = 0x00000001
DIR_MNTSTATUS_TRIGGER = 0x00000002
ATTR_DIR_ALLOCSIZE = 0x00000008
ATTR_DIR_IOBLOCKSIZE = 0x00000010
ATTR_DIR_DATALENGTH = 0x00000020
ATTR_DIR_ALL_ATTRS = 0x0000003F

ATTR_DIR_VALIDMASK = 0x0000003F
ATTR_DIR_SETMASK = 0x00000000

ATTR_FILE_LINKCOUNT = 0x00000001
ATTR_FILE_TOTALSIZE = 0x00000002
ATTR_FILE_ALLOCSIZE = 0x00000004
ATTR_FILE_IOBLOCKSIZE = 0x00000008
ATTR_FILE_DEVTYPE = 0x00000020
ATTR_FILE_FORKCOUNT = 0x00000080
ATTR_FILE_FORKLIST = 0x00000100
ATTR_FILE_DATALENGTH = 0x00000200
ATTR_FILE_DATAALLOCSIZE = 0x00000400
ATTR_FILE_RSRCLENGTH = 0x00001000
ATTR_FILE_RSRCALLOCSIZE = 0x00002000
ATTR_FILE_ALL_ATTRS = 0x000037FF

ATTR_FILE_VALIDMASK = 0x000037FF
ATTR_FILE_SETMASK = 0x00000020

# These are deprecated
ATTR_FORK_TOTALSIZE = 0x00000001
ATTR_FORK_ALLOCSIZE = 0x00000002
ATTR_FORK_ALL_ATTRS = 0x00000003

# These go in the fork attribute field
ATTR_CMNEXT_RELPATH = 0x00000004
ATTR_CMNEXT_PRIVATESIZE = 0x00000008
ATTR_CMNEXT_LINKID = 0x0000010
ATTR_CMNEXT_NOFIRMLINKPATH = 0x00000020
ATTR_CMNEXT_REALDEVID = 0x00000040
ATTR_CMNEXT_REALFSID = 0x00000080
ATTR_CMNEXT_CLONEID = 0x00000100
ATTR_CMNEXT_EXT_FLAGS = 0x00000200
ATTR_CMNEXT_RECURSIVE_GENCOUNT = 0x00000400
ATTR_CMNEXT_ALL_ATTRS = 0x000007FC

ATTR_CMNEXT_VALIDMASK = 0x000007FC
ATTR_CMNEXT_SETMASK = 0x00000000

ATTR_FORK_VALIDMASK = 0x00000003
ATTR_FORK_SETMASK = 0x00000000

# These can't be used
ATTR_CMN_NAMEDATTRCOUNT = 0x00080000
ATTR_CMN_NAMEDATTRLIST = 0x00100000
ATTR_FILE_CLUMPSIZE = 0x00000010
ATTR_FILE_FILETYPE = 0x00000040
ATTR_FILE_DATAEXTENTS = 0x00000800
ATTR_FILE_RSRCEXTENTS = 0x00004000


class attrlist(Structure):
    _fields_ = [
        ("bitmapcount", c_ushort),
        ("reserved", c_ushort),
        ("commonattr", c_uint),
        ("volattr", c_uint),
        ("dirattr", c_uint),
        ("fileattr", c_uint),
        ("forkattr", c_uint),
    ]


class attribute_set_t(Structure):
    _fields_ = [
        ("commonattr", c_uint),
        ("volattr", c_uint),
        ("dirattr", c_uint),
        ("fileattr", c_uint),
        ("forkattr", c_uint),
    ]


class fsobj_id_t(Structure):
    _fields_ = [
        ("fid_objno", c_uint),
        ("fid_generation", c_uint),
    ]


class timespec(Structure):
    _fields_ = [
        ("tv_sec", c_long),
        ("tv_nsec", c_long),
    ]


class attrreference_t(Structure):
    _fields_ = [
        ("attr_dataoffset", c_int),
        ("attr_length", c_uint),
    ]


class fsid_t(Structure):
    _fields_ = [
        ("val", c_uint * 2),
    ]


class guid_t(Structure):
    _fields_ = [
        ("g_guid", c_byte * 16),
    ]


class kauth_ace(Structure):
    _fields_ = [
        ("ace_applicable", guid_t),
        ("ace_flags", c_uint),
    ]


class kauth_acl(Structure):
    _fields_ = [
        ("acl_entrycount", c_uint),
        ("acl_flags", c_uint),
        ("acl_ace", kauth_ace * 128),
    ]


class kauth_filesec(Structure):
    _fields_ = [
        ("fsec_magic", c_uint),
        ("fsec_owner", guid_t),
        ("fsec_group", guid_t),
        ("fsec_acl", kauth_acl),
    ]


class diskextent(Structure):
    _fields_ = [
        ("startblock", c_uint),
        ("blockcount", c_uint),
    ]


OSType = c_uint
UInt16 = c_ushort
SInt16 = c_short
SInt32 = c_int


class Point(Structure):
    _fields_ = [
        ("x", SInt16),
        ("y", SInt16),
    ]


class Rect(Structure):
    _fields_ = [
        ("x", SInt16),
        ("y", SInt16),
        ("w", SInt16),
        ("h", SInt16),
    ]


class FileInfo(Structure):
    _fields_ = [
        ("fileType", OSType),
        ("fileCreator", OSType),
        ("finderFlags", UInt16),
        ("location", Point),
        ("reservedField", UInt16),
        ("reserved1", SInt16 * 4),
        ("extendedFinderFlags", UInt16),
        ("reserved2", SInt16),
        ("putAwayFolderID", SInt32),
    ]


class FolderInfo(Structure):
    _fields_ = [
        ("windowBounds", Rect),
        ("finderFlags", UInt16),
        ("location", Point),
        ("reservedField", UInt16),
        ("scrollPosition", Point),
        ("reserved1", SInt32),
        ("extendedFinderFlags", UInt16),
        ("reserved2", SInt16),
        ("putAwayFolderID", SInt32),
    ]


class FinderInfo(Union):
    _fields_ = [
        ("fileInfo", FileInfo),
        ("folderInfo", FolderInfo),
    ]


extentrecord = diskextent * 8
vol_capabilities_set_t = c_uint * 4


class vol_capabilities_attr_t(Structure):
    _fields_ = [
        ("capabilities", vol_capabilities_set_t),
        ("valid", vol_capabilities_set_t),
    ]


class vol_attributes_attr_t(Structure):
    _fields_ = [
        ("validattr", attribute_set_t),
        ("nativeattr", attribute_set_t),
    ]


dev_t = c_uint

fsobj_type_t = c_uint

VNON = 0
VREG = 1
VDIR = 2
VBLK = 3
VCHR = 4
VLNK = 5
VSOCK = 6
VFIFO = 7
VBAD = 8
VSTR = 9
VCPLX = 10

fsobj_tag_t = c_uint

VT_NON = 0
VT_UFS = 1
VT_NFS = 2
VT_MFS = 3
VT_MSDOSFS = 4
VT_LFS = 5
VT_LOFS = 6
VT_FDESC = 7
VT_PORTAL = 8
VT_NULL = 9
VT_UMAP = 10
VT_KERNFS = 11
VT_PROCFS = 12
VT_AFS = 13
VT_ISOFS = 14
VT_UNION = 15
VT_HFS = 16
VT_ZFS = 17
VT_DEVFS = 18
VT_WEBDAV = 19
VT_UDF = 20
VT_AFP = 21
VT_CDDA = 22
VT_CIFS = 23
VT_OTHER = 24

fsfile_type_t = c_uint
fsvolid_t = c_uint
text_encoding_t = c_uint
uid_t = c_uint
gid_t = c_uint
int32_t = c_int
uint32_t = c_uint
int64_t = c_longlong
uint64_t = c_ulonglong
off_t = c_long
size_t = c_ulong
uuid_t = c_byte * 16

NAME_MAX = 255
PATH_MAX = 1024
FSTYPE_MAX = 16


class struct_statfs(Structure):
    _fields_ = [
        ("f_bsize", uint32_t),
        ("f_iosize", int32_t),
        ("f_blocks", uint64_t),
        ("f_bfree", uint64_t),
        ("f_bavail", uint64_t),
        ("f_files", uint64_t),
        ("f_ffree", uint64_t),
        ("f_fsid", fsid_t),
        ("f_owner", uid_t),
        ("f_type", uint32_t),
        ("f_flags", uint32_t),
        ("f_fssubtype", uint32_t),
        ("f_fstypename", c_char * FSTYPE_MAX),
        ("f_mntonname", c_char * PATH_MAX),
        ("f_mntfromname", c_char * PATH_MAX),
        ("f_flags_ext", uint32_t),
        ("f_reserved", uint32_t * 7),
    ]


# Calculate the maximum number of bytes required for the attribute buffer
_attr_info = (
    # Common attributes
    (0, ATTR_CMN_RETURNED_ATTRS, sizeof(attribute_set_t)),
    (0, ATTR_CMN_NAME, sizeof(attrreference_t) + NAME_MAX * 3 + 1),
    (0, ATTR_CMN_DEVID, sizeof(dev_t)),
    (0, ATTR_CMN_FSID, sizeof(fsid_t)),
    (0, ATTR_CMN_OBJTYPE, sizeof(fsobj_type_t)),
    (0, ATTR_CMN_OBJTAG, sizeof(fsobj_tag_t)),
    (0, ATTR_CMN_OBJID, sizeof(fsobj_id_t)),
    (0, ATTR_CMN_OBJPERMANENTID, sizeof(fsobj_id_t)),
    (0, ATTR_CMN_PAROBJID, sizeof(fsobj_id_t)),
    (0, ATTR_CMN_SCRIPT, sizeof(text_encoding_t)),
    (0, ATTR_CMN_CRTIME, sizeof(timespec)),
    (0, ATTR_CMN_MODTIME, sizeof(timespec)),
    (0, ATTR_CMN_CHGTIME, sizeof(timespec)),
    (0, ATTR_CMN_ACCTIME, sizeof(timespec)),
    (0, ATTR_CMN_BKUPTIME, sizeof(timespec)),
    (0, ATTR_CMN_FNDRINFO, sizeof(FinderInfo)),
    (0, ATTR_CMN_OWNERID, sizeof(uid_t)),
    (0, ATTR_CMN_GRPID, sizeof(gid_t)),
    (0, ATTR_CMN_ACCESSMASK, sizeof(uint32_t)),
    (0, ATTR_CMN_NAMEDATTRCOUNT, None),
    (0, ATTR_CMN_NAMEDATTRLIST, None),
    (0, ATTR_CMN_FLAGS, sizeof(uint32_t)),
    (0, ATTR_CMN_GEN_COUNT, sizeof(uint32_t)),
    (0, ATTR_CMN_DOCUMENT_ID, sizeof(uint32_t)),
    (0, ATTR_CMN_USERACCESS, sizeof(uint32_t)),
    (0, ATTR_CMN_EXTENDED_SECURITY, sizeof(attrreference_t) + sizeof(kauth_filesec)),
    (0, ATTR_CMN_UUID, sizeof(guid_t)),
    (0, ATTR_CMN_GRPUUID, sizeof(guid_t)),
    (0, ATTR_CMN_FILEID, sizeof(uint64_t)),
    (0, ATTR_CMN_PARENTID, sizeof(uint64_t)),
    (0, ATTR_CMN_FULLPATH, sizeof(attrreference_t) + PATH_MAX),
    (0, ATTR_CMN_ADDEDTIME, sizeof(timespec)),
    (0, ATTR_CMN_DATA_PROTECT_FLAGS, sizeof(uint32_t)),
    # Volume attributes
    (1, ATTR_VOL_FSTYPE, sizeof(uint32_t)),
    (1, ATTR_VOL_SIGNATURE, sizeof(uint32_t)),
    (1, ATTR_VOL_SIZE, sizeof(off_t)),
    (1, ATTR_VOL_SPACEFREE, sizeof(off_t)),
    (1, ATTR_VOL_SPACEAVAIL, sizeof(off_t)),
    (1, ATTR_VOL_MINALLOCATION, sizeof(off_t)),
    (1, ATTR_VOL_ALLOCATIONCLUMP, sizeof(off_t)),
    (1, ATTR_VOL_IOBLOCKSIZE, sizeof(uint32_t)),
    (1, ATTR_VOL_OBJCOUNT, sizeof(uint32_t)),
    (1, ATTR_VOL_FILECOUNT, sizeof(uint32_t)),
    (1, ATTR_VOL_DIRCOUNT, sizeof(uint32_t)),
    (1, ATTR_VOL_MAXOBJCOUNT, sizeof(uint32_t)),
    (1, ATTR_VOL_MOUNTPOINT, sizeof(attrreference_t) + PATH_MAX),
    (1, ATTR_VOL_NAME, sizeof(attrreference_t) + NAME_MAX + 1),
    (1, ATTR_VOL_MOUNTFLAGS, sizeof(uint32_t)),
    (1, ATTR_VOL_MOUNTEDDEVICE, sizeof(attrreference_t) + PATH_MAX),
    (1, ATTR_VOL_ENCODINGSUSED, sizeof(c_ulonglong)),
    (1, ATTR_VOL_CAPABILITIES, sizeof(vol_capabilities_attr_t)),
    (1, ATTR_VOL_UUID, sizeof(uuid_t)),
    (1, ATTR_VOL_QUOTA_SIZE, sizeof(off_t)),
    (1, ATTR_VOL_RESERVED_SIZE, sizeof(off_t)),
    (1, ATTR_VOL_ATTRIBUTES, sizeof(vol_attributes_attr_t)),
    # Directory attributes
    (2, ATTR_DIR_LINKCOUNT, sizeof(uint32_t)),
    (2, ATTR_DIR_ENTRYCOUNT, sizeof(uint32_t)),
    (2, ATTR_DIR_MOUNTSTATUS, sizeof(uint32_t)),
    (2, ATTR_DIR_ALLOCSIZE, sizeof(off_t)),
    (2, ATTR_DIR_IOBLOCKSIZE, sizeof(uint32_t)),
    (2, ATTR_DIR_DATALENGTH, sizeof(off_t)),
    # File attributes
    (3, ATTR_FILE_LINKCOUNT, sizeof(uint32_t)),
    (3, ATTR_FILE_TOTALSIZE, sizeof(off_t)),
    (3, ATTR_FILE_ALLOCSIZE, sizeof(off_t)),
    (3, ATTR_FILE_IOBLOCKSIZE, sizeof(uint32_t)),
    (3, ATTR_FILE_CLUMPSIZE, sizeof(uint32_t)),
    (3, ATTR_FILE_DEVTYPE, sizeof(uint32_t)),
    (3, ATTR_FILE_FILETYPE, sizeof(uint32_t)),
    (3, ATTR_FILE_FORKCOUNT, sizeof(uint32_t)),
    (3, ATTR_FILE_FORKLIST, None),
    (3, ATTR_FILE_DATALENGTH, sizeof(off_t)),
    (3, ATTR_FILE_DATAALLOCSIZE, sizeof(off_t)),
    (3, ATTR_FILE_DATAEXTENTS, sizeof(extentrecord)),
    (3, ATTR_FILE_RSRCLENGTH, sizeof(off_t)),
    (3, ATTR_FILE_RSRCALLOCSIZE, sizeof(off_t)),
    (3, ATTR_FILE_RSRCEXTENTS, sizeof(extentrecord)),
    # Fork attributes
    (4, ATTR_FORK_TOTALSIZE, sizeof(off_t)),
    (4, ATTR_FORK_ALLOCSIZE, sizeof(off_t)),
    # Extended common attributes
    (4, ATTR_CMNEXT_RELPATH, sizeof(attrreference_t) + PATH_MAX),
    (4, ATTR_CMNEXT_PRIVATESIZE, sizeof(off_t)),
    (4, ATTR_CMNEXT_LINKID, sizeof(uint64_t)),
    (4, ATTR_CMNEXT_NOFIRMLINKPATH, sizeof(attrreference_t) + PATH_MAX),
    (4, ATTR_CMNEXT_REALDEVID, sizeof(dev_t)),
    (4, ATTR_CMNEXT_REALFSID, sizeof(fsid_t)),
    (4, ATTR_CMNEXT_CLONEID, sizeof(uint64_t)),
    (4, ATTR_CMNEXT_EXT_FLAGS, sizeof(uint64_t)),
)


def _attrbuf_size(attrs):
    size = 4
    for entry in _attr_info:
        if attrs[entry[0]] & entry[1]:
            if entry[2] is None:
                raise ValueError(
                    "Unsupported attribute (%u, %x)" % (entry[0], entry[1])
                )
            size += entry[2]
    return size


_getattrlist = libc.getattrlist
_getattrlist.argtypes = [c_char_p, POINTER(attrlist), c_void_p, c_ulong, c_ulong]
_getattrlist.restype = c_int

_fgetattrlist = libc.fgetattrlist
_fgetattrlist.argtypes = [c_int, POINTER(attrlist), c_void_p, c_ulong, c_ulong]
_fgetattrlist.restype = c_int

try:
    _statfs = libc["statfs$INODE64"]
except (KeyError, AttributeError):
    _statfs = libc["statfs"]

_statfs.argtypes = [
    c_char_p,
    POINTER(struct_statfs),
]
_statfs.restype = c_int

try:
    _fstatfs = libc["fstatfs$INODE64"]
except (KeyError, AttributeError):
    _fstatfs = libc["fstatfs"]

_fstatfs.argtypes = [
    c_int,
    POINTER(struct_statfs),
]
_fstatfs.restype = c_int


def _datetime_from_timespec(ts):
    td = datetime.timedelta(seconds=ts.tv_sec + 1.0e-9 * ts.tv_nsec)
    return unix_epoch + td


def _decode_utf8_nul(sz):
    nul = sz.find(b"\0")
    if nul > -1:
        sz = sz[:nul]
    return sz.decode("utf-8")


def _decode_attrlist_result(buf, attrs, options):
    result = []

    assert len(buf) >= 4
    total_size = uint32_t.from_buffer(buf, 0).value
    assert total_size <= len(buf)

    offset = 4

    # Common attributes
    if attrs[0] & ATTR_CMN_RETURNED_ATTRS:
        a = attribute_set_t.from_buffer(buf, offset)
        result.append(a)
        offset += sizeof(attribute_set_t)
        if not (options & FSOPT_PACK_INVAL_ATTRS):
            attrs = [a.commonattr, a.volattr, a.dirattr, a.fileattr, a.forkattr]
    if attrs[0] & ATTR_CMN_NAME:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        name = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(name)
    if attrs[0] & ATTR_CMN_DEVID:
        a = dev_t.from_buffer(buf, offset)
        offset += sizeof(dev_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_FSID:
        a = fsid_t.from_buffer(buf, offset)
        offset += sizeof(fsid_t)
        result.append(a)
    if attrs[0] & ATTR_CMN_OBJTYPE:
        a = fsobj_type_t.from_buffer(buf, offset)
        offset += sizeof(fsobj_type_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_OBJTAG:
        a = fsobj_tag_t.from_buffer(buf, offset)
        offset += sizeof(fsobj_tag_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_OBJID:
        a = fsobj_id_t.from_buffer(buf, offset)
        offset += sizeof(fsobj_id_t)
        result.append(a)
    if attrs[0] & ATTR_CMN_OBJPERMANENTID:
        a = fsobj_id_t.from_buffer(buf, offset)
        offset += sizeof(fsobj_id_t)
        result.append(a)
    if attrs[0] & ATTR_CMN_PAROBJID:
        a = fsobj_id_t.from_buffer(buf, offset)
        offset += sizeof(fsobj_id_t)
        result.append(a)
    if attrs[0] & ATTR_CMN_SCRIPT:
        a = text_encoding_t.from_buffer(buf, offset)
        offset += sizeof(text_encoding_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_CRTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_MODTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_CHGTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_ACCTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_BKUPTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_FNDRINFO:
        a = FinderInfo.from_buffer(buf, offset)
        offset += sizeof(FinderInfo)
        result.append(a)
    if attrs[0] & ATTR_CMN_OWNERID:
        a = uid_t.from_buffer(buf, offset)
        offset += sizeof(uid_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_GRPID:
        a = gid_t.from_buffer(buf, offset)
        offset += sizeof(gid_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_ACCESSMASK:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_FLAGS:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_GEN_COUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_DOCUMENT_ID:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_USERACCESS:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_EXTENDED_SECURITY:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        offset += sizeof(attrreference_t)
        ec = uint32_t.from_buffer(buf, ofs + 36).value

        class kauth_acl(Structure):
            _fields_ = [
                ("acl_entrycount", c_uint),
                ("acl_flags", c_uint),
                ("acl_ace", kauth_ace * ec),
            ]

        class kauth_filesec(Structure):
            _fields_ = [
                ("fsec_magic", c_uint),
                ("fsec_owner", guid_t),
                ("fsec_group", guid_t),
                ("fsec_acl", kauth_acl),
            ]

        a = kauth_filesec.from_buffer(buf, ofs)
        result.append(a)
    if attrs[0] & ATTR_CMN_UUID:
        result.append(uuid.UUID(bytes=buf[offset : offset + 16]))
        offset += sizeof(guid_t)
    if attrs[0] & ATTR_CMN_GRPUUID:
        result.append(uuid.UUID(bytes=buf[offset : offset + 16]))
        offset += sizeof(guid_t)
    if attrs[0] & ATTR_CMN_FILEID:
        a = uint64_t.from_buffer(buf, offset)
        offset += sizeof(uint64_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_PARENTID:
        a = uint64_t.from_buffer(buf, offset)
        offset += sizeof(uint64_t)
        result.append(a.value)
    if attrs[0] & ATTR_CMN_FULLPATH:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        path = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(path)
    if attrs[0] & ATTR_CMN_ADDEDTIME:
        a = timespec.from_buffer(buf, offset)
        offset += sizeof(timespec)
        result.append(_datetime_from_timespec(a))
    if attrs[0] & ATTR_CMN_DATA_PROTECT_FLAGS:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)

    # Volume attributes
    if attrs[1] & ATTR_VOL_FSTYPE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_SIGNATURE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_SIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_SPACEFREE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_SPACEAVAIL:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_MINALLOCATION:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_ALLOCATIONCLUMP:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_IOBLOCKSIZE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_OBJCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_FILECOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_DIRCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_MAXOBJCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_MOUNTPOINT:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        path = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(path)
    if attrs[1] & ATTR_VOL_NAME:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        name = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(name)
    if attrs[1] & ATTR_VOL_MOUNTFLAGS:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_MOUNTEDDEVICE:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        path = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(path)
    if attrs[1] & ATTR_VOL_ENCODINGSUSED:
        a = c_ulonglong.from_buffer(buf, offset)
        offset += sizeof(c_ulonglong)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_CAPABILITIES:
        a = vol_capabilities_attr_t.from_buffer(buf, offset)
        offset += sizeof(vol_capabilities_attr_t)
        result.append(a)
    if attrs[1] & ATTR_VOL_UUID:
        result.append(uuid.UUID(bytes=buf[offset : offset + 16]))
        offset += sizeof(uuid_t)
    if attrs[1] & ATTR_VOL_QUOTA_SIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_RESERVED_SIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[1] & ATTR_VOL_ATTRIBUTES:
        a = vol_attributes_attr_t.from_buffer(buf, offset)
        offset += sizeof(vol_attributes_attr_t)
        result.append(a)

    # Directory attributes
    if attrs[2] & ATTR_DIR_LINKCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[2] & ATTR_DIR_ENTRYCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[2] & ATTR_DIR_MOUNTSTATUS:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[2] & ATTR_DIR_ALLOCSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[2] & ATTR_DIR_IOBLOCKSIZE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[2] & ATTR_DIR_DATALENGTH:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)

    # File attributes
    if attrs[3] & ATTR_FILE_LINKCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_TOTALSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_ALLOCSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_IOBLOCKSIZE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_CLUMPSIZE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_DEVTYPE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_FILETYPE:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_FORKCOUNT:
        a = uint32_t.from_buffer(buf, offset)
        offset += sizeof(uint32_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_DATALENGTH:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_DATAALLOCSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_DATAEXTENTS:
        a = extentrecord.from_buffer(buf, offset)
        offset += sizeof(extentrecord)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_RSRCLENGTH:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_RSRCALLOCSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[3] & ATTR_FILE_RSRCEXTENTS:
        a = extentrecord.from_buffer(buf, offset)
        offset += sizeof(extentrecord)
        result.append(a.value)

    # Fork attributes
    if attrs[4] & ATTR_FORK_TOTALSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[4] & ATTR_FORK_ALLOCSIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)

    # Extended common attributes
    if attrs[4] & ATTR_CMNEXT_RELPATH:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        path = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(path)
    if attrs[4] & ATTR_CMNEXT_PRIVATESIZE:
        a = off_t.from_buffer(buf, offset)
        offset += sizeof(off_t)
        result.append(a.value)
    if attrs[4] & ATTR_CMNEXT_LINKID:
        a = uint64_t.from_buffer(buf, offset)
        offset += sizeof(uint64_t)
        result.append(a.value)
    if attrs[4] & ATTR_CMNEXT_NOFIRMLINKPATH:
        a = attrreference_t.from_buffer(buf, offset)
        ofs = offset + a.attr_dataoffset
        path = _decode_utf8_nul(buf[ofs : ofs + a.attr_length])
        offset += sizeof(attrreference_t)
        result.append(path)
    if attrs[4] & ATTR_CMNEXT_REALDEVID:
        a = dev_t.from_buffer(buf, offset)
        offset += sizeof(dev_t)
        result.append(a.value)
    if attrs[4] & ATTR_CMNEXT_REALFSID:
        a = fsid_t.from_buffer(buf, offset)
        offset += sizeof(fsid_t)
        result.append(a.value)
    if attrs[4] & ATTR_CMNEXT_CLONEID:
        a = uint64_t.from_buffer(buf, offset)
        offset += sizeof(uint64_t)
        result.append(a.value)
    if attrs[4] & ATTR_CMNEXT_EXT_FLAGS:
        a = uint64_t.from_buffer(buf, offset)
        offset += sizeof(uint64_t)
        result.append(a.value)

    return result


# Sadly, ctypes.get_errno() seems not to work
__error = libc.__error
__error.restype = POINTER(c_int)


def _get_errno():
    return __error().contents.value


def getattrlist(path, attrs, options):
    if not isinstance(path, bytes):
        path = path.encode("utf-8")
    attrs = list(attrs)
    if attrs[1]:
        attrs[1] |= ATTR_VOL_INFO
    alist = attrlist(
        bitmapcount=5,
        commonattr=attrs[0],
        volattr=attrs[1],
        dirattr=attrs[2],
        fileattr=attrs[3],
        forkattr=attrs[4],
    )

    bufsize = _attrbuf_size(attrs)
    buf = create_string_buffer(bufsize)

    ret = _getattrlist(
        path, byref(alist), buf, bufsize, options | FSOPT_REPORT_FULLSIZE
    )

    if ret < 0:
        err = _get_errno()
        raise OSError(err, os.strerror(err), path)

    return _decode_attrlist_result(buf, attrs, options)


def fgetattrlist(fd, attrs, options):
    if hasattr(fd, "fileno"):
        fd = fd.fileno()
    attrs = list(attrs)
    if attrs[1]:
        attrs[1] |= ATTR_VOL_INFO
    alist = attrlist(
        bitmapcount=5,
        commonattr=attrs[0],
        volattr=attrs[1],
        dirattr=attrs[2],
        fileattr=attrs[3],
        forkattr=attrs[4],
    )

    bufsize = _attrbuf_size(attrs)
    buf = create_string_buffer(bufsize)

    ret = _fgetattrlist(fd, byref(alist), buf, bufsize, options | FSOPT_REPORT_FULLSIZE)

    if ret < 0:
        err = _get_errno()
        raise OSError(err, os.strerror(err))

    return _decode_attrlist_result(buf, attrs, options)


def statfs(path):
    if not isinstance(path, bytes):
        path = path.encode("utf-8")
    result = struct_statfs()
    ret = _statfs(path, byref(result))
    if ret < 0:
        err = _get_errno()
        raise OSError(err, os.strerror(err), path)
    return result


def fstatfs(fd):
    if hasattr(fd, "fileno"):
        fd = fd.fileno()
    result = struct_statfs()
    ret = _fstatfs(fd, byref(result))
    if ret < 0:
        err = _get_errno()
        raise OSError(err, os.strerror(err))
    return result



import binascii
import bisect
import os
import struct


class BuddyError(Exception):
    pass


class Block:
    def __init__(self, allocator, offset, size):
        self._allocator = allocator
        self._offset = offset
        self._size = size
        self._value = bytearray(allocator.read(offset, size))
        self._pos = 0
        self._dirty = False

    def __len__(self):
        return self._size

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        if self._dirty:
            self.flush()

    def flush(self):
        if self._dirty:
            self._dirty = False
            self._allocator.write(self._offset, self._value)

    def invalidate(self):
        self._dirty = False

    def zero_fill(self):
        len = self._size - self._pos
        zeroes = b"\0" * len
        self._value[self._pos : self._size] = zeroes
        self._dirty = True

    def tell(self):
        return self._pos

    def seek(self, pos, whence=os.SEEK_SET):
        if whence == os.SEEK_CUR:
            pos += self._pos
        elif whence == os.SEEK_END:
            pos = self._size - pos

        if pos < 0 or pos > self._size:
            raise ValueError("Seek out of range in Block instance")

        self._pos = pos

    def read(self, size_or_format):
        if isinstance(size_or_format, (str, bytes)):
            size = struct.calcsize(size_or_format)
            fmt = size_or_format
        else:
            size = size_or_format
            fmt = None

        if self._size - self._pos < size:
            raise BuddyError("Unable to read %lu bytes in block" % size)

        data = self._value[self._pos : self._pos + size]
        self._pos += size

        if fmt is not None:
            if isinstance(data, bytearray):
                return struct.unpack_from(fmt, bytes(data))
            else:
                return struct.unpack(fmt, data)
        else:
            return data

    def write(self, data_or_format, *args):
        if len(args):
            data = struct.pack(data_or_format, *args)
        else:
            data = data_or_format

        if self._pos + len(data) > self._size:
            raise ValueError("Attempt to write past end of Block")

        self._value[self._pos : self._pos + len(data)] = data
        self._pos += len(data)

        self._dirty = True

    def insert(self, data_or_format, *args):
        if len(args):
            data = struct.pack(data_or_format, *args)
        else:
            data = data_or_format

        del self._value[-len(data) :]
        self._value[self._pos : self._pos] = data
        self._pos += len(data)

        self._dirty = True

    def delete(self, size):
        if self._pos + size > self._size:
            raise ValueError("Attempt to delete past end of Block")
        del self._value[self._pos : self._pos + size]
        self._value += b"\0" * size
        self._dirty = True

    def __str__(self):
        return binascii.b2a_hex(self._value)


class Allocator:
    def __init__(self, the_file):
        self._file = the_file
        self._dirty = False

        self._file.seek(0)

        # Read the header
        (magic1, magic2, offset, size, offset2, self._unknown1) = self.read(
            -4, ">I4sIII16s"
        )

        if magic2 != b"Bud1" or magic1 != 1:
            raise BuddyError("Not a buddy file")

        if offset != offset2:
            raise BuddyError("Root addresses differ")

        self._root = Block(self, offset, size)

        # Read the block offsets
        count, self._unknown2 = self._root.read(">II")
        self._offsets = []
        c = (count + 255) & ~255
        while c:
            self._offsets += self._root.read(">256I")
            c -= 256
        self._offsets = self._offsets[:count]

        # Read the TOC
        self._toc = {}
        count = self._root.read(">I")[0]
        for n in range(count):
            nlen = self._root.read("B")[0]
            name = bytes(self._root.read(nlen))
            value = self._root.read(">I")[0]
            self._toc[name] = value

        # Read the free lists
        self._free = []
        for n in range(32):
            count = self._root.read(">I")
            self._free.append(list(self._root.read(">%uI" % count)))

    @classmethod
    def open(cls, file_or_name, mode="r+"):
        if isinstance(file_or_name, str):
            if "b" not in mode:
                mode = mode[:1] + "b" + mode[1:]
            f = open(file_or_name, mode)
        else:
            f = file_or_name

        if "w" in mode:
            # Create an empty file in this case
            f.truncate()

            # An empty root block needs 1264 bytes:
            #
            #     0  4       offset count
            #     4  4       unknown
            #     8  4       root block offset (2048)
            #    12  255 * 4 padding (offsets are in multiples of 256)
            #  1032  4       toc count (0)
            #  1036  228     free list
            #  total 1264

            # The free list will contain the following:
            #
            #     0  5 * 4   no blocks of width less than 5
            #    20  6 * 8   1 block each of widths 5 to 10
            #    68  4       no blocks of width 11 (allocated for the root)
            #    72  19 * 8  1 block each of widths 12 to 30
            #   224  4       no blocks of width 31
            # total  228
            #
            # (The reason for this layout is that we allocate 2**5 bytes for
            #  the header, which splits the initial 2GB region into every size
            #  below 2**31, including *two* blocks of size 2**5, one of which
            #  we take.  The root block itself then needs a block of size
            #  2**11.  Conveniently, each of these initial blocks will be
            #  located at offset 2**n where n is its width.)

            # Write the header
            header = struct.pack(
                b">I4sIII16s",
                1,
                b"Bud1",
                2048,
                1264,
                2048,
                b"\x00\x00\x10\x0c"
                b"\x00\x00\x00\x87"
                b"\x00\x00\x20\x0b"
                b"\x00\x00\x00\x00",
            )
            f.write(header)
            f.write(b"\0" * 2016)

            # Write the root block
            free_list = [struct.pack(b">5I", 0, 0, 0, 0, 0)]
            for n in range(5, 11):
                free_list.append(struct.pack(b">II", 1, 2**n))
            free_list.append(struct.pack(b">I", 0))
            for n in range(12, 31):
                free_list.append(struct.pack(b">II", 1, 2**n))
            free_list.append(struct.pack(b">I", 0))

            root = b"".join(
                [
                    struct.pack(b">III", 1, 0, 2048 | 5),
                    struct.pack(b">I", 0) * 255,
                    struct.pack(b">I", 0),
                ]
                + free_list
            )
            f.write(root)

        return Allocator(f)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.flush()
        self._file.close()

    def flush(self):
        if self._dirty:
            size = self._root_block_size()
            self.allocate(size, 0)
            with self.get_block(0) as rblk:
                self._write_root_block_into(rblk)

            addr = self._offsets[0]
            offset = addr & ~0x1F
            size = 1 << (addr & 0x1F)

            self._file.seek(0, os.SEEK_SET)
            self._file.write(
                struct.pack(
                    b">I4sIII16s", 1, b"Bud1", offset, size, offset, self._unknown1
                )
            )

            self._dirty = False

        self._file.flush()

    def read(self, offset, size_or_format):
        """Read data at `offset', or raise an exception.

        `size_or_format' may either be a byte count, in which case we
        return raw data, or a format string for `struct.unpack', in
        which case we work out the size and unpack the data before
        returning it.
        """
        # N.B. There is a fixed offset of four bytes(!)
        self._file.seek(offset + 4, os.SEEK_SET)

        if isinstance(size_or_format, str):
            size = struct.calcsize(size_or_format)
            fmt = size_or_format
        else:
            size = size_or_format
            fmt = None

        ret = self._file.read(size)
        if len(ret) < size:
            ret += b"\0" * (size - len(ret))

        if fmt is not None:
            if isinstance(ret, bytearray):
                ret = struct.unpack_from(fmt, bytes(ret))
            else:
                ret = struct.unpack(fmt, ret)

        return ret

    def write(self, offset, data_or_format, *args):
        """Write data at `offset', or raise an exception.

        `data_or_format' may either be the data to write, or a format
        string for `struct.pack', in which case we pack the additional
        arguments and write the resulting data.
        """
        # N.B. There is a fixed offset of four bytes(!)
        self._file.seek(offset + 4, os.SEEK_SET)

        if len(args):
            data = struct.pack(data_or_format, *args)
        else:
            data = data_or_format

        self._file.write(data)

    def get_block(self, block):
        try:
            addr = self._offsets[block]
        except IndexError:
            return None

        offset = addr & ~0x1F
        size = 1 << (addr & 0x1F)

        return Block(self, offset, size)

    def _root_block_size(self):
        """Return the number of bytes required by the root block."""
        # Offsets
        size = 8
        size += 4 * ((len(self._offsets) + 255) & ~255)

        # TOC
        size += 4
        size += sum([5 + len(s) for s in self._toc])

        # Free list
        size += sum([4 + 4 * len(fl) for fl in self._free])

        return size

    def _write_root_block_into(self, block):
        # Offsets
        block.write(">II", len(self._offsets), self._unknown2)
        block.write(">%uI" % len(self._offsets), *self._offsets)
        extra = len(self._offsets) & 255
        if extra:
            block.write(b"\0\0\0\0" * (256 - extra))

        # TOC
        keys = list(self._toc.keys())
        keys.sort()

        block.write(">I", len(keys))
        for k in keys:
            block.write("B", len(k))
            block.write(k)
            block.write(">I", self._toc[k])

        # Free list
        for w, f in enumerate(self._free):
            block.write(">I", len(f))
            if len(f):
                block.write(">%uI" % len(f), *f)

    def _buddy(self, offset, width):
        f = self._free[width]
        b = offset ^ (1 << width)

        try:
            ndx = f.index(b)
        except ValueError:
            ndx = None

        return (f, b, ndx)

    def _release(self, offset, width):
        # Coalesce
        while True:
            f, b, ndx = self._buddy(offset, width)

            if ndx is None:
                break

            offset &= b
            width += 1
            del f[ndx]

        # Add to the list
        bisect.insort(f, offset)

        # Mark as dirty
        self._dirty = True

    def _alloc(self, width):
        w = width
        while not self._free[w]:
            w += 1
        while w > width:
            offset = self._free[w].pop(0)
            w -= 1
            self._free[w] = [offset, offset ^ (1 << w)]
        self._dirty = True
        return self._free[width].pop(0)

    def allocate(self, bytes, block=None):
        """Allocate or reallocate a block such that it has space for at least
        `bytes' bytes."""
        if block is None:
            # Find the first unused block
            try:
                block = self._offsets.index(0)
            except ValueError:
                block = len(self._offsets)
                self._offsets.append(0)

        # Compute block width
        width = max(bytes.bit_length(), 5)

        addr = self._offsets[block]
        offset = addr & ~0x1F

        if addr:
            blkwidth = addr & 0x1F
            if blkwidth == width:
                return block
            self._release(offset, width)
            self._offsets[block] = 0

        offset = self._alloc(width)
        self._offsets[block] = offset | width
        return block

    def release(self, block):
        addr = self._offsets[block]

        if addr:
            width = addr & 0x1F
            offset = addr & ~0x1F
            self._release(offset, width)

        if block == len(self._offsets):
            del self._offsets[block]
        else:
            self._offsets[block] = 0

    def __len__(self):
        return len(self._toc)

    def __getitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("Keys must be of string type")
        if not isinstance(key, bytes):
            key = key.encode("latin_1")
        return self._toc[key]

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Keys must be of string type")
        if not isinstance(key, bytes):
            key = key.encode("latin_1")
        self._toc[key] = value
        self._dirty = True

    def __delitem__(self, key):
        if not isinstance(key, str):
            raise TypeError("Keys must be of string type")
        if not isinstance(key, bytes):
            key = key.encode("latin_1")
        del self._toc[key]
        self._dirty = True

    def iterkeys(self):
        return self._toc.keys()

    def keys(self):
        return self._toc.keys()

    def __iter__(self):
        return self._toc.keys()

    def __contains__(self, key):
        return key in self._toc



import plistlib
import struct

import mac_alias

from . import buddy


class ILocCodec:
    @staticmethod
    def encode(point):
        return struct.pack(b">IIII", point[0], point[1], 0xFFFFFFFF, 0xFFFF0000)

    @staticmethod
    def decode(bytesData):
        if isinstance(bytesData, bytearray):
            x, y = struct.unpack_from(b">II", bytes(bytesData[:8]))
        else:
            x, y = struct.unpack(b">II", bytesData[:8])
        return (x, y)


class PlistCodec:
    @staticmethod
    def encode(plist):
        return plistlib.dumps(plist, fmt=plistlib.FMT_BINARY)

    @staticmethod
    def decode(bytes):
        return plistlib.loads(bytes)


class BookmarkCodec:
    @staticmethod
    def encode(bmk):
        return bmk.to_bytes()

    @staticmethod
    def decode(bytes):
        return mac_alias.Bookmark.from_bytes(bytes)


# This list tells the code how to decode particular kinds of entry in the
# .DS_Store file.  This is really a convenience, and we currently only
# support a tiny subset of the possible entry types.
codecs = {
    b"Iloc": ILocCodec,
    b"bwsp": PlistCodec,
    b"lsvp": PlistCodec,
    b"lsvP": PlistCodec,
    b"icvp": PlistCodec,
    b"pBBk": BookmarkCodec,
}


class DSStoreEntry:
    """Holds the data from an entry in a ``.DS_Store`` file.  Note that this is
    not meant to represent the entry itself---i.e. if you change the type or
    value, your changes will *not* be reflected in the underlying file.

    If you want to make a change, you should either use the
    :class:`DSStore` object's :meth:`DSStore.insert` method (which will
    replace a key if it already exists), or the mapping access mode for
    :class:`DSStore` (often simpler anyway).
    """

    def __init__(self, filename, code, typecode, value=None):
        if str != bytes and type(filename) == bytes:
            filename = filename.decode("utf-8")

        if not isinstance(code, bytes):
            code = code.encode("latin_1")

        self.filename = filename
        self.code = code
        self.type = typecode
        self.value = value

    @classmethod
    def read(cls, block):
        """Read a ``.DS_Store`` entry from the containing Block."""
        # First read the filename
        nlen = block.read(b">I")[0]
        filename = block.read(2 * nlen).decode("utf-16be")

        # Next, read the code and type
        code, typecode = block.read(b">4s4s")

        # Finally, read the data
        if typecode == b"bool":
            value = block.read(b">?")[0]
        elif typecode == b"long" or typecode == b"shor":
            value = block.read(b">I")[0]
        elif typecode == b"blob":
            vlen = block.read(b">I")[0]
            value = block.read(vlen)

            codec = codecs.get(code, None)
            if codec:
                value = codec.decode(value)
                typecode = codec
        elif typecode == b"ustr":
            vlen = block.read(b">I")[0]
            value = block.read(2 * vlen).decode("utf-16be")
        elif typecode == b"type":
            value = block.read(b">4s")[0]
        elif typecode == b"comp" or typecode == b"dutc":
            value = block.read(b">Q")[0]
        else:
            raise ValueError('Unknown type code "%s"' % typecode)

        return DSStoreEntry(filename, code, typecode, value)

    def __lt__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()
        return sfl < ofl or (self.filename == other.filename and self.code < other.code)

    def __le__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()
        return sfl < ofl or (sfl == ofl and self.code <= other.code)

    def __eq__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()
        return sfl == ofl and self.code == other.code

    def __ne__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()
        return sfl != ofl or self.code != other.code

    def __gt__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()

        selfCode = self.code
        if str != bytes and type(selfCode) is bytes:
            selfCode = selfCode.decode("utf-8")
        otherCode = other.code
        if str != bytes and type(otherCode) is bytes:
            otherCode = otherCode.decode("utf-8")

        return sfl > ofl or (sfl == ofl and selfCode > otherCode)

    def __ge__(self, other):
        if not isinstance(other, DSStoreEntry):
            raise TypeError("Can only compare against other DSStoreEntry objects")
        sfl = self.filename.lower()
        ofl = other.filename.lower()
        return sfl > ofl or (sfl == ofl and self.code >= other.code)

    def byte_length(self):
        """Compute the length of this entry, in bytes."""
        utf16 = self.filename.encode("utf-16be")
        length = 4 + len(utf16) + 8

        if isinstance(self.type, str):
            entry_type = self.type.encode("latin_1")
            value = self.value
        elif isinstance(self.type, (bytes, str)):
            entry_type = self.type
            value = self.value
        else:
            entry_type = b"blob"
            value = self.type.encode(self.value)

        if entry_type == b"bool":
            length += 1
        elif entry_type == b"long" or entry_type == b"shor":
            length += 4
        elif entry_type == b"blob":
            length += 4 + len(value)
        elif entry_type == b"ustr":
            utf16 = value.encode("utf-16be")
            length += 4 + len(utf16)
        elif entry_type == b"type":
            length += 4
        elif entry_type == b"comp" or entry_type == b"dutc":
            length += 8
        else:
            raise ValueError('Unknown type code "%s"' % entry_type)

        return length

    def write(self, block, insert=False):
        """Write this entry to the specified Block."""
        if insert:
            w = block.insert
        else:
            w = block.write

        if isinstance(self.type, str):
            entry_type = self.type.encode("latin_1")
            value = self.value
        elif isinstance(self.type, (bytes, str)):
            entry_type = self.type
            value = self.value
        else:
            entry_type = b"blob"
            value = self.type.encode(self.value)

        utf16 = self.filename.encode("utf-16be")
        w(b">I", len(utf16) // 2)
        w(utf16)
        w(b">4s4s", self.code, entry_type)

        if entry_type == b"bool":
            w(b">?", value)
        elif entry_type == b"long" or entry_type == b"shor":
            w(b">I", value)
        elif entry_type == b"blob":
            w(b">I", len(value))
            w(value)
        elif entry_type == b"ustr":
            utf16 = value.encode("utf-16be")
            w(b">I", len(utf16) // 2)
            w(utf16)
        elif entry_type == b"type":
            if isinstance(value, str):
                value = value.encode("latin_1")
            w(b">4s", value)
        elif entry_type == b"comp" or entry_type == b"dutc":
            w(b">Q", value)
        else:
            raise ValueError('Unknown type code "%s"' % entry_type)

    def __repr__(self):
        return f"<{self.filename} {self.code}>"


class DSStore:
    """Python interface to a ``.DS_Store`` file.  Works by manipulating the
    file on the disk---so this code will work with ``.DS_Store`` files for.

    *very* large directories.

    A :class:`DSStore` object can be used as if it was a mapping, e.g.::

      d['foobar.dat']['Iloc']

    will fetch the "Iloc" record for "foobar.dat", or raise :class:`KeyError` if
    there is no such record.  If used in this manner, the :class:`DSStore` object
    will return (type, value) tuples, unless the type is "blob" and the module
    knows how to decode it.

    Currently, we know how to decode "Iloc", "bwsp", "lsvp", "lsvP" and "icvp"
    blobs.  "Iloc" decodes to an (x, y) tuple, while the others are all decoded
    using ``biplist`` or ``plistlib`` depending on Python version.

    Assignment also works, e.g.::

      d['foobar.dat']['note'] = ('ustr', u'Hello World!')

    as does deletion with ``del``::

      del d['foobar.dat']['note']

    This is usually going to be the most convenient interface, though
    occasionally (for instance when creating a new ``.DS_Store`` file) you
    may wish to drop down to using :class:`DSStoreEntry` objects directly.
    """

    def __init__(self, store):
        self._store = store
        self._superblk = self._store["DSDB"]
        with self._get_block(self._superblk) as s:
            (
                self._rootnode,
                self._levels,
                self._records,
                self._nodes,
                self._page_size,
            ) = s.read(b">IIIII")
        self._min_usage = 2 * self._page_size // 3
        self._dirty = False

    @classmethod
    def open(cls, file_or_name, mode="r+", initial_entries=None):
        """Open a ``.DS_Store`` file; pass either a Python file object, or a
        filename in the ``file_or_name`` argument and a file access mode in the
        ``mode`` argument.

        If you are creating a new file using the "w" or "w+" modes, you
        may also specify a list of entries with which to initialise the
        file.
        """
        store = buddy.Allocator.open(file_or_name, mode)

        if mode == "w" or mode == "w+":
            superblk = store.allocate(20)
            store["DSDB"] = superblk
            page_size = 4096

            if not initial_entries:
                root = store.allocate(page_size)

                with store.get_block(root) as rootblk:
                    rootblk.zero_fill()

                with store.get_block(superblk) as s:
                    s.write(b">IIIII", root, 0, 0, 1, page_size)
            else:
                # Make sure they're in sorted order
                initial_entries = list(initial_entries)
                initial_entries.sort()

                # Construct the tree
                current_level = initial_entries
                next_level = []
                levels = []
                ptr_size = 0
                node_count = 0
                while True:
                    total = 8
                    nodes = []
                    node = []
                    for e in current_level:
                        new_total = total + ptr_size + e.byte_length()
                        if new_total > page_size:
                            nodes.append(node)
                            next_level.append(e)
                            total = 8
                            node = []
                        else:
                            total = new_total
                            node.append(e)
                    if node:
                        nodes.append(node)

                    node_count += len(nodes)
                    levels.append(nodes)

                    if len(nodes) == 1:
                        break

                    current_level = next_level
                    next_level = []
                    ptr_size = 4

                # Allocate nodes
                ptrs = [store.allocate(page_size) for n in range(node_count)]

                # Generate nodes
                pointers = []
                prev_pointers = None
                for level in levels:
                    ppndx = 0
                    lptrs = ptrs[-len(level) :]
                    del ptrs[-len(level) :]
                    for node in level:
                        ndx = lptrs.pop(0)
                        if prev_pointers is None:
                            with store.get_block(ndx) as block:
                                block.write(b">II", 0, len(node))
                                for e in node:
                                    e.write(block)
                        else:
                            next_node = prev_pointers[ppndx + len(node)]
                            node_ptrs = prev_pointers[ppndx : ppndx + len(node)]

                            with store.get_block(ndx) as block:
                                block.write(b">II", next_node, len(node))
                                for ptr, e in zip(node_ptrs, node):
                                    block.write(b">I", ptr)
                                    e.write(block)

                        pointers.append(ndx)
                    prev_pointers = pointers
                    pointers = []

                root = prev_pointers[0]

                with store.get_block(superblk) as s:
                    s.write(
                        b">IIIII",
                        root,
                        len(levels),
                        len(initial_entries),
                        node_count,
                        page_size,
                    )

        return DSStore(store)

    def _get_block(self, number):
        return self._store.get_block(number)

    def flush(self):
        """Flush any dirty data back to the file."""
        if self._dirty:
            self._dirty = False

            with self._get_block(self._superblk) as s:
                s.write(
                    b">IIIII",
                    self._rootnode,
                    self._levels,
                    self._records,
                    self._nodes,
                    self._page_size,
                )
        self._store.flush()

    def close(self):
        """Flush dirty data and close the underlying file."""
        self.flush()
        self._store.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    # Internal B-Tree nodes look like this:
    #
    # [ next | count | (ptr0 | rec0) | (ptr1 | rec1) ... (ptrN | recN) ]

    # Leaf nodes look like this:
    #
    # [ 0 | count | rec0 | rec1 ... recN ]

    # Iterate over the tree, starting at `node'
    def _traverse(self, node):
        if node is None:
            node = self._rootnode
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")
            if next_node:
                for n in range(count):
                    ptr = block.read(b">I")[0]
                    for e in self._traverse(ptr):
                        yield e
                    e = DSStoreEntry.read(block)
                    yield e
                for e in self._traverse(next_node):
                    yield e
            else:
                for n in range(count):
                    e = DSStoreEntry.read(block)
                    yield e

    # Display the data in `node'
    def _dump_node(self, node):
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")
            print("next: %u\ncount: %u\n" % (next_node, count))
            for n in range(count):
                if next_node:
                    ptr = block.read(b">I")[0]
                    print("%8u " % ptr, end=" ")
                else:
                    print("         ", end=" ")
                e = DSStoreEntry.read(block)
                print(e, " (%u)" % e.byte_length())
            print("used: %u" % block.tell())

    # Display the data in the super block
    def _dump_super(self):
        print(
            "root: %u\nlevels: %u\nrecords: %u\nnodes: %u\npage-size: %u"
            % (
                self._rootnode,
                self._levels,
                self._records,
                self._nodes,
                self._page_size,
            )
        )

    # Splits entries across two blocks, returning one pivot
    #
    # Tries to balance the block usage across the two as best it can
    def _split2(self, blocks, entries, pointers, before, internal):
        left_block = blocks[0]
        right_block = blocks[1]

        count = len(entries)

        # Find the feasible splits
        best_split = None
        best_diff = None
        total = before[count]

        if 8 + total <= self._page_size:
            # We can use a *single* node for this
            best_split = count
        else:
            # Split into two nodes
            for n in range(1, count - 1):
                left_size = 8 + before[n]
                right_size = 8 + total - before[n + 1]

                if left_size > self._page_size:
                    break
                if right_size > self._page_size:
                    continue

                diff = abs(left_size - right_size)

                if best_split is None or diff < best_diff:
                    best_split = n
                    best_diff = diff

        if best_split is None:
            return None

        # Write the nodes
        left_block.seek(0)
        if internal:
            next_node = pointers[best_split]
        else:
            next_node = 0
        left_block.write(b">II", next_node, best_split)

        for n in range(best_split):
            if internal:
                left_block.write(b">I", pointers[n])
            entries[n].write(left_block)

        left_block.zero_fill()

        if best_split == count:
            return []

        right_block.seek(0)
        if internal:
            next_node = pointers[count]
        else:
            next_node = 0
        right_block.write(b">II", next_node, count - best_split - 1)

        for n in range(best_split + 1, count):
            if internal:
                right_block.write(b">I", pointers[n])
            entries[n].write(right_block)

        right_block.zero_fill()

        pivot = entries[best_split]

        return [pivot]

    def _split(self, node, entry, right_ptr=0):
        self._nodes += 1
        self._dirty = True
        new_right = self._store.allocate(self._page_size)
        with self._get_block(node) as block, self._get_block(new_right) as right_block:

            # First, measure and extract all the elements
            entry_size = entry.byte_length()
            # ?? entry_pos = None
            next_node, count = block.read(b">II")
            if next_node:
                entry_size += 4
            pointers = []
            entries = []
            before = []
            total = 0
            for n in range(count):
                pos = block.tell()
                if next_node:
                    ptr = block.read(b">I")[0]
                    pointers.append(ptr)
                e = DSStoreEntry.read(block)
                if e > entry:
                    # ?? entry_pos = n
                    entries.append(entry)
                    pointers.append(right_ptr)
                    before.append(total)
                    total += entry_size
                entries.append(e)
                before.append(total)
                total += block.tell() - pos
            before.append(total)
            if next_node:
                pointers.append(next_node)

            pivot = self._split2(
                [block, right_block], entries, pointers, before, bool(next_node)
            )[0]

            self._records += 1
            self._nodes += 1
            self._dirty = True

        return (pivot, new_right)

    # Allocate a new root node containing the element `pivot' and the pointers
    # `left' and `right'
    def _new_root(self, left, pivot, right):
        new_root = self._store.allocate(self._page_size)
        with self._get_block(new_root) as block:
            block.write(b">III", right, 1, left)
            pivot.write(block)
        self._rootnode = new_root
        self._levels += 1
        self._nodes += 1
        self._dirty = True

    # Insert an entry into an inner node; `path' is the path from the root
    # to `node', not including `node' itself.  `right_ptr' is the new node
    # pointer (inserted to the RIGHT of `entry')
    def _insert_inner(self, path, node, entry, right_ptr):
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")
            insert_pos = None
            insert_ndx = None
            n = 0
            while n < count:
                pos = block.tell()
                ptr = block.read(b">I")[0]
                e = DSStoreEntry.read(block)
                if e == entry:
                    if n == count - 1:
                        right_ptr = next_node
                        next_node = ptr
                        block.seek(pos)
                    else:
                        right_ptr = block.read(b">I")[0]
                        block.seek(pos + 4)
                    insert_pos = pos
                    insert_ndx = n
                    block.delete(e.byte_length() + 4)
                    count -= 1
                    self._records += 1
                    self._dirty = True
                    continue
                elif insert_pos is None and e > entry:
                    insert_pos = pos
                    insert_ndx = n
                n += 1
            if insert_pos is None:
                insert_pos = block.tell()
                insert_ndx = count
            remaining = self._page_size - block.tell()

            if remaining < entry.byte_length() + 4:
                pivot, new_right = self._split(node, entry, right_ptr)
                if path:
                    self._insert_inner(path[:-1], path[-1], pivot, new_right)
                else:
                    self._new_root(node, pivot, new_right)
            else:
                if insert_ndx == count:
                    block.seek(insert_pos)
                    block.write(b">I", next_node)
                    entry.write(block)
                    next_node = right_ptr
                else:
                    block.seek(insert_pos + 4)
                    entry.write(block, True)
                    block.insert(">I", right_ptr)
                block.seek(0)
                count += 1
                block.write(b">II", next_node, count)
                self._records += 1
                self._dirty = True

    # Insert `entry' into the leaf node `node'
    def _insert_leaf(self, path, node, entry):
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")
            insert_pos = None
            # ?? insert_ndx = None
            n = 0
            while n < count:
                pos = block.tell()
                e = DSStoreEntry.read(block)
                if e == entry:
                    insert_pos = pos
                    # ?? insert_ndx = n
                    block.seek(pos)
                    block.delete(e.byte_length())
                    count -= 1
                    self._records += 1
                    self._dirty = True
                    continue
                elif insert_pos is None and e > entry:
                    insert_pos = pos
                    # ?? insert_ndx = n
                n += 1
            if insert_pos is None:
                insert_pos = block.tell()
                # ?? insert_ndx = count
            remaining = self._page_size - block.tell()

            if remaining < entry.byte_length():
                pivot, new_right = self._split(node, entry)
                if path:
                    self._insert_inner(path[:-1], path[-1], pivot, new_right)
                else:
                    self._new_root(node, pivot, new_right)
            else:
                block.seek(insert_pos)
                entry.write(block, True)
                block.seek(0)
                count += 1
                block.write(b">II", next_node, count)
                self._records += 1
                self._dirty = True

    def insert(self, entry):
        """Insert ``entry`` (which should be a :class:`DSStoreEntry`) into the
        B-Tree."""
        path = []
        node = self._rootnode
        while True:
            with self._get_block(node) as block:
                next_node, count = block.read(b">II")
                if next_node:
                    for n in range(count):
                        ptr = block.read(b">I")[0]
                        e = DSStoreEntry.read(block)
                        if entry < e:
                            next_node = ptr
                            break
                        elif entry == e:
                            # If we find an existing entry the same, replace it
                            self._insert_inner(path, node, entry, None)
                            return
                    path.append(node)
                    node = next_node
                else:
                    self._insert_leaf(path, node, entry)
                    return

    # Return usage information for the specified `node'
    def _block_usage(self, node):
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")

            for n in range(count):
                if next_node:
                    block.read(b">I")[0]
                DSStoreEntry.read(block)

            used = block.tell()

        return (count, used)

    # Splits entries across three blocks, returning two pivots
    def _split3(self, blocks, entries, pointers, before, internal):
        count = len(entries)

        # Find the feasible splits
        best_split = None
        best_diff = None
        total = before[count]
        for n in range(1, count - 3):
            left_size = 8 + before[n]
            remaining = 16 + total - before[n + 1]

            if left_size > self._page_size:
                break
            if remaining > 2 * self._page_size:
                continue

            for m in range(n + 2, count - 1):
                mid_size = 8 + before[m] - before[n + 1]
                right_size = 8 + total - before[m + 1]

                if mid_size > self._page_size:
                    break
                if right_size > self._page_size:
                    continue

                diff = abs(left_size - mid_size) * abs(right_size - mid_size)

                if best_split is None or diff < best_diff:
                    best_split = (n, m, count)
                    best_diff = diff

        if best_split is None:
            return None

        # Write the nodes
        prev_split = -1
        for block, split in zip(blocks, best_split):
            block.seek(0)
            if internal:
                next_node = pointers[split]
            else:
                next_node = 0
            block.write(b">II", next_node, split)

            for n in range(prev_split + 1, split):
                if internal:
                    block.write(b">I", pointers[n])
                entries[n].write(block)

            block.zero_fill()

            prev_split = split

        return (entries[best_split[0]], entries[best_split[1]])

    # Extract all of the entries from the specified list of `blocks',
    # separating them by the specified `pivots'.  Also computes the
    # amount of space used before each entry.
    def _extract(self, blocks, pivots):
        pointers = []
        entries = []
        before = []
        total = 0
        ppivots = pivots + [None]
        for b, p in zip(blocks, ppivots):
            b.seek(0)
            next_node, count = b.read(b">II")
            for n in range(count):
                pos = b.tell()
                if next_node:
                    ptr = b.read(b">I")[0]
                    pointers.append(ptr)
                e = DSStoreEntry.read(b)
                entries.append(e)
                before.append(total)
                total += b.tell() - pos
            if next_node:
                pointers.append(next_node)
            if p:
                entries.append(p)
                before.append(total)
                total += p.byte_length()
                if next_node:
                    total += 4
        before.append(total)

        return (entries, pointers, before)

    # Rebalance the specified `node', whose path from the root is `path'.
    def _rebalance(self, path, node):
        # Can't rebalance the root
        if not path:
            return

        with self._get_block(node) as block:
            next_node, count = block.read(b">II")

            with self._get_block(path[-1]) as parent:
                # Find the left and right siblings and respective pivots
                parent_next, parent_count = parent.read(b">II")
                left_pos = None
                left_node = None
                left_pivot = None
                node_pos = None
                right_pos = None
                right_node = None
                right_pivot = None
                prev_e = prev_ptr = prev_pos = None
                for n in range(parent_count):
                    pos = parent.tell()
                    ptr = parent.read(b">I")[0]
                    e = DSStoreEntry.read(parent)

                    if ptr == node:
                        node_pos = pos
                        right_pivot = e
                        left_pos = prev_pos
                        left_pivot = prev_e
                        left_node = prev_ptr
                    elif prev_ptr == node:
                        right_node = ptr
                        right_pos = pos
                        break

                    prev_e = e
                    prev_ptr = ptr
                    prev_pos = pos

                if parent_next == node:
                    node_pos = parent.tell()
                    left_pos = prev_pos
                    left_pivot = prev_e
                    left_node = prev_ptr
                elif right_node is None:
                    right_node = parent_next
                    right_pos = parent.tell()

                _ = parent.tell()

            if left_node and right_node:
                with self._get_block(left_node) as left, self._get_block(
                    right_node
                ) as right:
                    blocks = [left, block, right]
                    pivots = [left_pivot, right_pivot]

                    entries, pointers, before = self._extract(blocks, pivots)

                    # If there's a chance that we could use two pages instead
                    # of three, go for it
                    pivots = self._split2(
                        blocks, entries, pointers, before, bool(next_node)
                    )
                    if pivots is None:
                        ptrs = [left_node, node, right_node]
                        pivots = self._split3(
                            blocks, entries, pointers, before, bool(next_node)
                        )
                    else:
                        if pivots:
                            ptrs = [left_node, node]
                        else:
                            ptrs = [left_node]
                            self._store.release(node)
                            self._nodes -= 1
                            node = left_node
                        self._store.release(right_node)
                        self._nodes -= 1
                        self._dirty = True

                    # Remove the pivots from the parent
                    with self._get_block(path[-1]) as parent:
                        if right_node == parent_next:
                            parent.seek(left_pos)
                            parent.delete(right_pos - left_pos)
                            parent_next = left_node
                        else:
                            parent.seek(left_pos + 4)
                            parent.delete(right_pos - left_pos)
                        parent.seek(0)
                        parent_count -= 2
                        parent.write(b">II", parent_next, parent_count)
                        self._records -= 2

                    # Replace with those in pivots
                    for e, rp in zip(pivots, ptrs[1:]):
                        self._insert_inner(path[:-1], path[-1], e, rp)
            elif left_node:
                with self._get_block(left_node) as left:
                    blocks = [left, block]
                    pivots = [left_pivot]

                    entries, pointers, before = self._extract(blocks, pivots)

                    pivots = self._split2(
                        blocks, entries, pointers, before, bool(next_node)
                    )

                    # Remove the pivot from the parent
                    with self._get_block(path[-1]) as parent:
                        if node == parent_next:
                            parent.seek(left_pos)
                            parent.delete(node_pos - left_pos)
                            parent_next = left_node
                        else:
                            parent.seek(left_pos + 4)
                            parent.delete(node_pos - left_pos)
                        parent.seek(0)
                        parent_count -= 1
                        parent.write(b">II", parent_next, parent_count)
                        self._records -= 1

                    # Replace the pivot
                    if pivots:
                        self._insert_inner(path[:-1], path[-1], pivots[0], node)
            elif right_node:
                with self._get_block(right_node) as right:
                    blocks = [block, right]
                    pivots = [right_pivot]

                    entries, pointers, before = self._extract(blocks, pivots)

                    pivots = self._split2(
                        blocks, entries, pointers, before, bool(next_node)
                    )

                    # Remove the pivot from the parent
                    with self._get_block(path[-1]) as parent:
                        if right_node == parent_next:
                            parent.seek(pos)
                            parent.delete(right_pos - node_pos)
                            parent_next = node
                        else:
                            parent.seek(pos + 4)
                            parent.delete(right_pos - node_pos)
                        parent.seek(0)
                        parent_count -= 1
                        parent.write(b">II", parent_next, parent_count)
                        self._records -= 1

                    # Replace the pivot
                    if pivots:
                        self._insert_inner(path[:-1], path[-1], pivots[0], right_node)

        if not path and not parent_count:
            self._store.release(path[-1])
            self._nodes -= 1
            self._dirty = True
            self._rootnode = node
        else:
            count, used = self._block_usage(path[-1])

            if used < self._page_size // 2:
                self._rebalance(path[:-1], path[-1])

    # Delete from the leaf node `node'.  `filename_lc' has already been
    # lower-cased.
    def _delete_leaf(self, node, filename_lc, code):
        found = False

        with self._get_block(node) as block:
            next_node, count = block.read(b">II")

            for n in range(count):
                pos = block.tell()
                e = DSStoreEntry.read(block)
                if e.filename.lower() == filename_lc and (
                    code is None or e.code == code
                ):
                    block.seek(pos)
                    block.delete(e.byte_length())
                    found = True

                    # This does not affect the loop; THIS IS NOT A BUG
                    count -= 1

                    self._records -= 1
                    self._dirty = True

            if found:
                used = block.tell()

                block.seek(0)
                block.write(b">II", next_node, count)

                return used < self._page_size // 2
            else:
                return False

    # Remove the largest entry from the subtree starting at `node' (with
    # path from root `path').  Returns a tuple (rebalance, entry) where
    # rebalance is either None if no rebalancing is required, or a
    # (path, node) tuple giving the details of the node to rebalance.
    def _take_largest(self, path, node):
        path = list(path)
        rebalance = None
        while True:
            with self._get_block(node) as block:
                next_node, count = block.read(b">II")

                if next_node:
                    path.append(node)
                    node = next_node
                    continue

                for n in range(count):
                    pos = block.tell()
                    e = DSStoreEntry.read(block)

                count -= 1
                block.seek(0)
                block.write(b">II", next_node, count)

                if pos < self._page_size // 2:
                    rebalance = (path, node)
                break

        return rebalance, e

    # Delete an entry from an inner node, `node'
    def _delete_inner(self, path, node, filename_lc, code):
        rebalance = False

        with self._get_block(node) as block:
            next_node, count = block.read(b">II")

            for n in range(count):
                pos = block.tell()
                ptr = block.read(b">I")[0]
                e = DSStoreEntry.read(block)
                if e.filename.lower() == filename_lc and (
                    code is None or e.code == code
                ):
                    # Take the largest from the left subtree
                    rebalance, largest = self._take_largest(path, ptr)

                    # Delete this entry
                    if n == count - 1:
                        right_ptr = next_node
                        next_node = ptr
                        block.seek(pos)
                    else:
                        right_ptr = block.read(b">I")[0]
                        block.seek(pos + 4)

                    block.delete(e.byte_length() + 4)

                    count -= 1
                    block.seek(0)
                    block.write(b">II", next_node, count)

                    self._records -= 1
                    self._dirty = True

                    break

        # Replace the pivot value
        self._insert_inner(path, node, largest, right_ptr)

        # Rebalance from the node we stole from
        if rebalance:
            self._rebalance(rebalance[0], rebalance[1])
            return True
        return False

    def delete(self, filename, code):
        """Delete an item, identified by ``filename`` and ``code`` from the
        B-Tree."""
        if isinstance(filename, DSStoreEntry):
            code = filename.code
            filename = filename.filename

        # If we're deleting *every* node for "filename", we must recurse
        if code is None:
            # TODO: Fix this so we can do bulk deletes
            raise ValueError("You must delete items individually.  Sorry")

        # Otherwise, we're deleting *one* specific node
        filename_lc = filename.lower()
        path = []
        node = self._rootnode
        while True:
            with self._get_block(node) as block:
                next_node, count = block.read(b">II")
                if next_node:
                    for n in range(count):
                        ptr = block.read(b">I")[0]
                        e = DSStoreEntry.read(block)
                        e_lc = e.filename.lower()
                        if filename_lc < e_lc or (
                            filename_lc == e_lc and code < e.code
                        ):
                            next_node = ptr
                            break
                        elif filename_lc == e_lc and code == e.code:
                            self._delete_inner(path, node, filename_lc, code)
                            return
                    path.append(node)
                    node = next_node
                else:
                    if self._delete_leaf(node, filename_lc, code):
                        self._rebalance(path, node)
                    return

    # Find implementation
    def _find(self, node, filename_lc, code=None):
        if code is not None and not isinstance(code, bytes):
            code = code.encode("latin_1")
        with self._get_block(node) as block:
            next_node, count = block.read(b">II")
            if next_node:
                for n in range(count):
                    ptr = block.read(b">I")[0]
                    e = DSStoreEntry.read(block)
                    if filename_lc < e.filename.lower():
                        for e in self._find(ptr, filename_lc, code):
                            yield e
                        return
                    elif filename_lc == e.filename.lower():
                        if code is None or (code and code < e.code):
                            for e in self._find(ptr, filename_lc, code):
                                yield e
                        if code is None or code == e.code:
                            yield e
                        elif code < e.code:
                            return
                for e in self._find(next_node, filename_lc, code):
                    yield e
            else:
                for n in range(count):
                    e = DSStoreEntry.read(block)
                    if filename_lc == e.filename.lower():
                        if code is None or code == e.code:
                            yield e
                        elif code < e.code:
                            return

    def find(self, filename, code=None):
        """Returns a generator that will iterate over matching entries in the
        B-Tree."""
        if isinstance(filename, DSStoreEntry):
            code = filename.code
            filename = filename.filename

        filename_lc = filename.lower()

        return self._find(self._rootnode, filename_lc, code)

    def __len__(self):
        return self._records

    def __iter__(self):
        return self._traverse(self._rootnode)

    class Partial:
        """This is used to implement indexing."""

        def __init__(self, store, filename):
            self._store = store
            self._filename = filename

        def __getitem__(self, code):
            if code is None:
                raise KeyError("no such key - [%s][None]" % self._filename)

            if not isinstance(code, bytes):
                code = code.encode("latin_1")

            try:
                item = next(self._store.find(self._filename, code))
            except StopIteration:
                raise KeyError(f"no such key - [{self._filename}][{code}]")

            if not isinstance(item.type, (bytes, str)):
                return item.value

            return (item.type, item.value)

        def __setitem__(self, code, value):
            if code is None:
                raise KeyError("bad key - [%s][None]" % self._filename)

            if not isinstance(code, bytes):
                code = code.encode("latin_1")

            codec = codecs.get(code, None)
            if codec:
                entry_type = codec
                entry_value = value
            else:
                entry_type = value[0]
                entry_value = value[1]

            self._store.insert(
                DSStoreEntry(self._filename, code, entry_type, entry_value)
            )

        def __delitem__(self, code):
            if code is None:
                raise KeyError("no such key - [%s][None]" % self._filename)

            self._store.delete(self._filename, code)

        def __iter__(self):
            yield from self._store.find(self._filename)

    def __getitem__(self, filename):
        return self.Partial(self, filename)



import datetime

ZERO = datetime.timedelta(0)


class UTC(datetime.tzinfo):
    def utcoffset(self, dt):
        return ZERO

    def dst(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"


utc = UTC()
mac_epoch = datetime.datetime(1904, 1, 1, 0, 0, 0, 0, utc)
unix_epoch = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, utc)
osx_epoch = datetime.datetime(2001, 1, 1, 0, 0, 0, 0, utc)



from conan import ConanFile
from conan.tools.qbs import Qbs, QbsProfile, QbsDeps

class Recipe(ConanFile):
    name = "conan-zlib"
    version = "1.0"

    exports_sources = "*.c", "*.qbs",
    settings = "os", "compiler", "arch", "build_type"

    requires = ["zlib/1.3.1"]

    def generate(self):
        profile = QbsProfile(self)
        profile.generate()
        deps = QbsDeps(self)
        deps.generate()

    def build(self):
        qbs = Qbs(self)
        qbs.resolve()
        qbs.build()

    def package(self):
        qbs = Qbs(self)
        qbs.install()



#  This file implements the Apple "bookmark" format, which is the replacement
#  for the old-fashioned alias format.  The details of this format were
#  reverse engineered; some things are still not entirely clear.
#
import datetime
import os
import struct
import sys
import uuid
from urllib.parse import urljoin

if sys.platform == "darwin":
    from . import osx

from .utils import osx_epoch

BMK_DATA_TYPE_MASK = 0xFFFFFF00
BMK_DATA_SUBTYPE_MASK = 0x000000FF

BMK_STRING = 0x0100
BMK_DATA = 0x0200
BMK_NUMBER = 0x0300
BMK_DATE = 0x0400
BMK_BOOLEAN = 0x0500
BMK_ARRAY = 0x0600
BMK_DICT = 0x0700
BMK_UUID = 0x0800
BMK_URL = 0x0900
BMK_NULL = 0x0A00

BMK_ST_ZERO = 0x0000
BMK_ST_ONE = 0x0001

BMK_BOOLEAN_ST_FALSE = 0x0000
BMK_BOOLEAN_ST_TRUE = 0x0001

# Subtypes for BMK_NUMBER are really CFNumberType values
kCFNumberSInt8Type = 1
kCFNumberSInt16Type = 2
kCFNumberSInt32Type = 3
kCFNumberSInt64Type = 4
kCFNumberFloat32Type = 5
kCFNumberFloat64Type = 6
kCFNumberCharType = 7
kCFNumberShortType = 8
kCFNumberIntType = 9
kCFNumberLongType = 10
kCFNumberLongLongType = 11
kCFNumberFloatType = 12
kCFNumberDoubleType = 13
kCFNumberCFIndexType = 14
kCFNumberNSIntegerType = 15
kCFNumberCGFloatType = 16

# Resource property flags (from CFURLPriv.h)
kCFURLResourceIsRegularFile = 0x00000001
kCFURLResourceIsDirectory = 0x00000002
kCFURLResourceIsSymbolicLink = 0x00000004
kCFURLResourceIsVolume = 0x00000008
kCFURLResourceIsPackage = 0x00000010
kCFURLResourceIsSystemImmutable = 0x00000020
kCFURLResourceIsUserImmutable = 0x00000040
kCFURLResourceIsHidden = 0x00000080
kCFURLResourceHasHiddenExtension = 0x00000100
kCFURLResourceIsApplication = 0x00000200
kCFURLResourceIsCompressed = 0x00000400
kCFURLResourceIsSystemCompressed = 0x00000400
kCFURLCanSetHiddenExtension = 0x00000800
kCFURLResourceIsReadable = 0x00001000
kCFURLResourceIsWriteable = 0x00002000
kCFURLResourceIsExecutable = 0x00004000
kCFURLIsAliasFile = 0x00008000
kCFURLIsMountTrigger = 0x00010000

# Volume property flags (from CFURLPriv.h)
kCFURLVolumeIsLocal = 0x1
kCFURLVolumeIsAutomount = 0x2
kCFURLVolumeDontBrowse = 0x4
kCFURLVolumeIsReadOnly = 0x8
kCFURLVolumeIsQuarantined = 0x10
kCFURLVolumeIsEjectable = 0x20
kCFURLVolumeIsRemovable = 0x40
kCFURLVolumeIsInternal = 0x80
kCFURLVolumeIsExternal = 0x100
kCFURLVolumeIsDiskImage = 0x200
kCFURLVolumeIsFileVault = 0x400
kCFURLVolumeIsLocaliDiskMirror = 0x800
kCFURLVolumeIsiPod = 0x1000
kCFURLVolumeIsiDisk = 0x2000
kCFURLVolumeIsCD = 0x4000
kCFURLVolumeIsDVD = 0x8000
kCFURLVolumeIsDeviceFileSystem = 0x10000
kCFURLVolumeSupportsPersistentIDs = 0x100000000
kCFURLVolumeSupportsSearchFS = 0x200000000
kCFURLVolumeSupportsExchange = 0x400000000
# reserved                                           0x800000000
kCFURLVolumeSupportsSymbolicLinks = 0x1000000000
kCFURLVolumeSupportsDenyModes = 0x2000000000
kCFURLVolumeSupportsCopyFile = 0x4000000000
kCFURLVolumeSupportsReadDirAttr = 0x8000000000
kCFURLVolumeSupportsJournaling = 0x10000000000
kCFURLVolumeSupportsRename = 0x20000000000
kCFURLVolumeSupportsFastStatFS = 0x40000000000
kCFURLVolumeSupportsCaseSensitiveNames = 0x80000000000
kCFURLVolumeSupportsCasePreservedNames = 0x100000000000
kCFURLVolumeSupportsFLock = 0x200000000000
kCFURLVolumeHasNoRootDirectoryTimes = 0x400000000000
kCFURLVolumeSupportsExtendedSecurity = 0x800000000000
kCFURLVolumeSupports2TBFileSize = 0x1000000000000
kCFURLVolumeSupportsHardLinks = 0x2000000000000
kCFURLVolumeSupportsMandatoryByteRangeLocks = 0x4000000000000
kCFURLVolumeSupportsPathFromID = 0x8000000000000
# reserved                                      0x10000000000000
kCFURLVolumeIsJournaling = 0x20000000000000
kCFURLVolumeSupportsSparseFiles = 0x40000000000000
kCFURLVolumeSupportsZeroRuns = 0x80000000000000
kCFURLVolumeSupportsVolumeSizes = 0x100000000000000
kCFURLVolumeSupportsRemoteEvents = 0x200000000000000
kCFURLVolumeSupportsHiddenFiles = 0x400000000000000
kCFURLVolumeSupportsDecmpFSCompression = 0x800000000000000
kCFURLVolumeHas64BitObjectIDs = 0x1000000000000000
kCFURLVolumePropertyFlagsAll = 0xFFFFFFFFFFFFFFFF

BMK_URL_ST_ABSOLUTE = 0x0001
BMK_URL_ST_RELATIVE = 0x0002

# Bookmark keys
kBookmarkURL = 0x1003  # A URL
kBookmarkPath = 0x1004  # Array of path components
kBookmarkCNIDPath = 0x1005  # Array of CNIDs
kBookmarkFileProperties = (
    0x1010  # (CFURL rp flags, CFURL rp flags asked for, 8 bytes NULL)
)
kBookmarkFileName = 0x1020
kBookmarkFileID = 0x1030
kBookmarkFileCreationDate = 0x1040
# = 0x1054   # ?
# = 0x1055   # ?
# = 0x1056   # ?
# = 0x1101   # ?
# = 0x1102   # ?
kBookmarkTOCPath = 0x2000  # A list of (TOC id, ?) pairs
kBookmarkVolumePath = 0x2002
kBookmarkVolumeURL = 0x2005
kBookmarkVolumeName = 0x2010
kBookmarkVolumeUUID = 0x2011  # Stored (perversely) as a string
kBookmarkVolumeSize = 0x2012
kBookmarkVolumeCreationDate = 0x2013
kBookmarkVolumeProperties = (
    0x2020  # (CFURL vp flags, CFURL vp flags asked for, 8 bytes NULL)
)
kBookmarkVolumeIsRoot = 0x2030  # True if volume is FS root
kBookmarkVolumeBookmark = 0x2040  # Embedded bookmark for disk image (TOC id)
kBookmarkVolumeMountPoint = 0x2050  # A URL
# = 0x2070
kBookmarkContainingFolder = 0xC001  # Index of containing folder in path
kBookmarkUserName = 0xC011  # User that created bookmark
kBookmarkUID = 0xC012  # UID that created bookmark
kBookmarkWasFileReference = 0xD001  # True if the URL was a file reference
kBookmarkCreationOptions = 0xD010
kBookmarkURLLengths = 0xE003  # See below
kBookmarkDisplayName = 0xF017
kBookmarkIconData = 0xF020
kBookmarkIconRef = 0xF021
kBookmarkTypeBindingData = 0xF022
kBookmarkCreationTime = 0xF030
kBookmarkSandboxRwExtension = 0xF080
kBookmarkSandboxRoExtension = 0xF081
kBookmarkAliasData = 0xFE00

# Alias for backwards compatibility
kBookmarkSecurityExtension = kBookmarkSandboxRwExtension

# kBookmarkURLLengths is an array that is set if the URL encoded by the
# bookmark had a base URL; in that case, each entry is the length of the
# base URL in question.  Thus a URL
#
#     file:///foo/bar/baz    blam/blat.html
#
# will result in [3, 2], while the URL
#
#     file:///foo    bar/baz    blam    blat.html
#
# would result in [1, 2, 1, 1]


class Data:
    def __init__(self, bytedata=None):
        #: The bytes, stored as a byte string
        self.bytes = bytes(bytedata)

    def __repr__(self):
        return "Data(%r)" % self.bytes


class URL:
    def __init__(self, base, rel=None):
        if rel is not None:
            #: The base URL, if any (a :class:`URL`)
            self.base = base
            #: The rest of the URL (a string)
            self.relative = rel
        else:
            self.base = None
            self.relative = base

    @property
    def absolute(self):
        """Return an absolute URL."""
        if self.base is None:
            return self.relative
        else:
            return urljoin(self.base.absolute, self.relative)

    def __repr__(self):
        return "URL(%r)" % self.absolute


class Bookmark:
    def __init__(self, tocs=None):
        if tocs is None:
            #: The TOCs for this Bookmark
            self.tocs = []
        else:
            self.tocs = tocs

    @classmethod
    def _get_item(cls, data, hdrsize, offset):
        offset += hdrsize
        if offset > len(data) - 8:
            raise ValueError("Offset out of range")

        length, typecode = struct.unpack(b"<II", data[offset : offset + 8])

        if len(data) - offset < 8 + length:
            raise ValueError("Data item truncated")

        databytes = data[offset + 8 : offset + 8 + length]

        dsubtype = typecode & BMK_DATA_SUBTYPE_MASK
        dtype = typecode & BMK_DATA_TYPE_MASK

        if dtype == BMK_STRING:
            return databytes.decode("utf-8")
        elif dtype == BMK_DATA:
            return Data(databytes)
        elif dtype == BMK_NUMBER:
            if dsubtype == kCFNumberSInt8Type:
                return ord(databytes[0])
            elif dsubtype == kCFNumberSInt16Type:
                return struct.unpack(b"<h", databytes)[0]
            elif dsubtype == kCFNumberSInt32Type:
                return struct.unpack(b"<i", databytes)[0]
            elif dsubtype == kCFNumberSInt64Type:
                return struct.unpack(b"<q", databytes)[0]
            elif dsubtype == kCFNumberFloat32Type:
                return struct.unpack(b"<f", databytes)[0]
            elif dsubtype == kCFNumberFloat64Type:
                return struct.unpack(b"<d", databytes)[0]
        elif dtype == BMK_DATE:
            # Yes, dates really are stored as *BIG-endian* doubles; everything
            # else is little-endian
            secs = datetime.timedelta(seconds=struct.unpack(b">d", databytes)[0])
            return osx_epoch + secs
        elif dtype == BMK_BOOLEAN:
            if dsubtype == BMK_BOOLEAN_ST_TRUE:
                return True
            elif dsubtype == BMK_BOOLEAN_ST_FALSE:
                return False
        elif dtype == BMK_UUID:
            return uuid.UUID(bytes=databytes)
        elif dtype == BMK_URL:
            if dsubtype == BMK_URL_ST_ABSOLUTE:
                return URL(databytes.decode("utf-8"))
            elif dsubtype == BMK_URL_ST_RELATIVE:
                baseoff, reloff = struct.unpack(b"<II", databytes)
                base = cls._get_item(data, hdrsize, baseoff)
                rel = cls._get_item(data, hdrsize, reloff)
                return URL(base, rel)
        elif dtype == BMK_ARRAY:
            result = []
            for aoff in range(offset + 8, offset + 8 + length, 4):
                (eltoff,) = struct.unpack(b"<I", data[aoff : aoff + 4])
                result.append(cls._get_item(data, hdrsize, eltoff))
            return result
        elif dtype == BMK_DICT:
            result = {}
            for eoff in range(offset + 8, offset + 8 + length, 8):
                keyoff, valoff = struct.unpack(b"<II", data[eoff : eoff + 8])
                key = cls._get_item(data, hdrsize, keyoff)
                val = cls._get_item(data, hdrsize, valoff)
                result[key] = val
            return result
        elif dtype == BMK_NULL:
            return None

        print("Unknown data type %08x" % typecode)
        return (typecode, databytes)

    @classmethod
    def from_bytes(cls, data):
        """Create a :class:`Bookmark` given byte data."""

        if len(data) < 16:
            raise ValueError("Not a bookmark file (too short)")

        if isinstance(data, bytearray):
            data = bytes(data)

        magic, size, dummy, hdrsize = struct.unpack(b"<4sIII", data[0:16])

        if magic not in (b"book", b"alis"):
            raise ValueError("Not a bookmark file (bad magic) %r" % magic)

        if hdrsize < 16:
            raise ValueError("Not a bookmark file (header size too short)")

        if hdrsize > size:
            raise ValueError("Not a bookmark file (header size too large)")

        if size != len(data):
            raise ValueError("Not a bookmark file (truncated)")

        (tocoffset,) = struct.unpack(b"<I", data[hdrsize : hdrsize + 4])

        tocs = []

        while tocoffset != 0:
            tocbase = hdrsize + tocoffset
            if (tocoffset > size - hdrsize) or (size - tocbase < 20):
                raise ValueError("TOC offset out of range")

            (tocsize, tocmagic, tocid, nexttoc, toccount) = struct.unpack(
                b"<IIIII", data[tocbase : tocbase + 20]
            )

            if tocmagic != 0xFFFFFFFE:
                break

            tocsize += 8

            if size - tocbase < tocsize:
                raise ValueError("TOC truncated")

            if tocsize < 12 * toccount:
                raise ValueError("TOC entries overrun TOC size")

            toc = {}
            for n in range(0, toccount):
                ebase = tocbase + 20 + 12 * n
                eid, eoffset, edummy = struct.unpack(b"<III", data[ebase : ebase + 12])

                if eid & 0x80000000:
                    eid = cls._get_item(data, hdrsize, eid & 0x7FFFFFFF)

                toc[eid] = cls._get_item(data, hdrsize, eoffset)

            tocs.append((tocid, toc))

            tocoffset = nexttoc

        return cls(tocs)

    def __getitem__(self, key):
        for tid, toc in self.tocs:
            if key in toc:
                return toc[key]
        raise KeyError("Key not found")

    def __setitem__(self, key, value):
        if len(self.tocs) == 0:
            self.tocs = [(1, {})]
        self.tocs[0][1][key] = value

    def get(self, key, default=None):
        """Lookup the value for a given key, returning a default if not
        present."""
        for tid, toc in self.tocs:
            if key in toc:
                return toc[key]
        return default

    @classmethod
    def _encode_item(cls, item, offset):
        if item is True:
            result = struct.pack(b"<II", 0, BMK_BOOLEAN | BMK_BOOLEAN_ST_TRUE)
        elif item is False:
            result = struct.pack(b"<II", 0, BMK_BOOLEAN | BMK_BOOLEAN_ST_FALSE)
        elif isinstance(item, str):
            encoded = item.encode("utf-8")
            result = (
                struct.pack(b"<II", len(encoded), BMK_STRING | BMK_ST_ONE) + encoded
            )
        elif isinstance(item, bytes):
            result = struct.pack(b"<II", len(item), BMK_STRING | BMK_ST_ONE) + item
        elif isinstance(item, Data):
            result = struct.pack(
                b"<II", len(item.bytes), BMK_DATA | BMK_ST_ONE
            ) + bytes(item.bytes)
        elif isinstance(item, bytearray):
            result = struct.pack(b"<II", len(item), BMK_DATA | BMK_ST_ONE) + bytes(item)
        elif isinstance(item, int):
            if item > -0x80000000 and item < 0x7FFFFFFF:
                result = struct.pack(b"<IIi", 4, BMK_NUMBER | kCFNumberSInt32Type, item)
            else:
                result = struct.pack(b"<IIq", 8, BMK_NUMBER | kCFNumberSInt64Type, item)
        elif isinstance(item, float):
            result = struct.pack(b"<IId", 8, BMK_NUMBER | kCFNumberFloat64Type, item)
        elif isinstance(item, datetime.datetime):
            secs = item - osx_epoch
            result = struct.pack(b"<II", 8, BMK_DATE | BMK_ST_ZERO) + struct.pack(
                b">d", float(secs.total_seconds())
            )

        elif isinstance(item, uuid.UUID):
            result = struct.pack(b"<II", 16, BMK_UUID | BMK_ST_ONE) + item.bytes
        elif isinstance(item, URL):
            if item.base:
                baseoff = offset + 16
                reloff, baseenc = cls._encode_item(item.base, baseoff)
                xoffset, relenc = cls._encode_item(item.relative, reloff)
                result = b"".join(
                    [
                        struct.pack(
                            b"<IIII", 8, BMK_URL | BMK_URL_ST_RELATIVE, baseoff, reloff
                        ),
                        baseenc,
                        relenc,
                    ]
                )
            else:
                encoded = item.relative.encode("utf-8")
                result = (
                    struct.pack(b"<II", len(encoded), BMK_URL | BMK_URL_ST_ABSOLUTE)
                    + encoded
                )
        elif isinstance(item, list):
            ioffset = offset + 8 + len(item) * 4
            result = [struct.pack(b"<II", len(item) * 4, BMK_ARRAY | BMK_ST_ONE)]
            enc = []
            for elt in item:
                result.append(struct.pack(b"<I", ioffset))
                ioffset, ienc = cls._encode_item(elt, ioffset)
                enc.append(ienc)
            result = b"".join(result + enc)
        elif isinstance(item, dict):
            ioffset = offset + 8 + len(item) * 8
            result = [struct.pack(b"<II", len(item) * 8, BMK_DICT | BMK_ST_ONE)]
            enc = []
            for k, v in item.items():
                result.append(struct.pack(b"<I", ioffset))
                ioffset, ienc = cls._encode_item(k, ioffset)
                enc.append(ienc)
                result.append(struct.pack(b"<I", ioffset))
                ioffset, ienc = cls._encode_item(v, ioffset)
                enc.append(ienc)
            result = b"".join(result + enc)
        elif item is None:
            result = struct.pack(b"<II", 0, BMK_NULL | BMK_ST_ONE)
        else:
            raise ValueError("Unknown item type when encoding: %s" % item)

        offset += len(result)

        # Pad to a multiple of 4 bytes
        if offset & 3:
            extra = 4 - (offset & 3)
            result += b"\0" * extra
            offset += extra

        return (offset, result)

    def to_bytes(self):
        """Convert this :class:`Bookmark` to a byte representation."""

        result = []
        tocs = []
        offset = 4  # For the offset to the first TOC

        # Generate the data and build the TOCs
        for tid, toc in self.tocs:
            entries = []

            for k, v in toc.items():
                if isinstance(k, str):
                    noffset = offset
                    voffset, enc = self._encode_item(k, offset)
                    result.append(enc)
                    offset, enc = self._encode_item(v, voffset)
                    result.append(enc)
                    entries.append((noffset | 0x80000000, voffset))
                else:
                    entries.append((k, offset))
                    offset, enc = self._encode_item(v, offset)
                    result.append(enc)

            # TOC entries must be sorted - CoreServicesInternal does a
            # binary search to find data
            entries.sort()

            tocs.append(
                (tid, b"".join([struct.pack(b"<III", k, o, 0) for k, o in entries]))
            )

        first_toc_offset = offset

        # Now generate the TOC headers
        for ndx, toc in enumerate(tocs):
            tid, data = toc
            if ndx == len(tocs) - 1:
                next_offset = 0
            else:
                next_offset = offset + 20 + len(data)

            result.append(
                struct.pack(
                    b"<IIIII",
                    len(data) - 8,
                    0xFFFFFFFE,
                    tid,
                    next_offset,
                    len(data) // 12,
                )
            )
            result.append(data)

            offset += 20 + len(data)

        # Finally, add the header (and the first TOC offset, which isn't part
        # of the header, but goes just after it)
        header = struct.pack(
            b"<4sIIIQQQQI",
            b"book",
            offset + 48,
            0x10040000,
            48,
            0,
            0,
            0,
            0,
            first_toc_offset,
        )

        result.insert(0, header)

        return b"".join(result)

    @classmethod
    def for_file(cls, path):
        """Construct a :class:`Bookmark` for a given file."""

        # Find the filesystem
        st = osx.statfs(path)
        vol_path = st.f_mntonname.decode("utf-8")

        # Grab its attributes
        attrs = [
            osx.ATTR_CMN_CRTIME,
            osx.ATTR_VOL_SIZE | osx.ATTR_VOL_NAME | osx.ATTR_VOL_UUID,
            0,
            0,
            0,
        ]
        volinfo = osx.getattrlist(vol_path, attrs, 0)

        vol_crtime = volinfo[0]
        vol_size = volinfo[1]
        vol_name = volinfo[2]
        vol_uuid = volinfo[3]

        # Also grab various attributes of the file
        attrs = [
            (osx.ATTR_CMN_OBJTYPE | osx.ATTR_CMN_CRTIME | osx.ATTR_CMN_FILEID),
            0,
            0,
            0,
            0,
        ]
        info = osx.getattrlist(path, attrs, osx.FSOPT_NOFOLLOW)

        cnid = info[2]
        crtime = info[1]

        if info[0] == osx.VREG:
            flags = kCFURLResourceIsRegularFile
        elif info[0] == osx.VDIR:
            flags = kCFURLResourceIsDirectory
        elif info[0] == osx.VLNK:
            flags = kCFURLResourceIsSymbolicLink
        else:
            flags = kCFURLResourceIsRegularFile

        dirname, filename = os.path.split(path)

        relcount = 0
        if not os.path.isabs(dirname):
            curdir = os.getcwd()
            head, tail = os.path.split(curdir)
            relcount = 0
            while head and tail:
                relcount += 1
                head, tail = os.path.split(head)
            dirname = os.path.join(curdir, dirname)

        # ?? foldername = os.path.basename(dirname)

        rel_path = os.path.relpath(path, vol_path)

        # Build the path arrays
        name_path = []
        cnid_path = []
        head, tail = os.path.split(rel_path)
        if not tail:
            head, tail = os.path.split(head)
        while head or tail:
            if head:
                attrs = [osx.ATTR_CMN_FILEID, 0, 0, 0, 0]
                info = osx.getattrlist(os.path.join(vol_path, head), attrs, 0)
                cnid_path.insert(0, info[0])
                head, tail = os.path.split(head)
                name_path.insert(0, tail)
            else:
                head, tail = os.path.split(head)
        name_path.append(filename)
        cnid_path.append(cnid)

        url_lengths = [relcount, len(name_path) - relcount]

        fileprops = Data(struct.pack(b"<QQQ", flags, 0x0F, 0))
        volprops = Data(
            struct.pack(
                b"<QQQ",
                0x81 | kCFURLVolumeSupportsPersistentIDs,
                0x13EF | kCFURLVolumeSupportsPersistentIDs,
                0,
            )
        )

        toc = {
            kBookmarkPath: name_path,
            kBookmarkCNIDPath: cnid_path,
            kBookmarkFileCreationDate: crtime,
            kBookmarkFileProperties: fileprops,
            kBookmarkContainingFolder: len(name_path) - 2,
            kBookmarkVolumePath: vol_path,
            kBookmarkVolumeIsRoot: vol_path == "/",
            kBookmarkVolumeURL: URL("file://" + vol_path),
            kBookmarkVolumeName: vol_name,
            kBookmarkVolumeSize: vol_size,
            kBookmarkVolumeCreationDate: vol_crtime,
            kBookmarkVolumeUUID: str(vol_uuid).upper(),
            kBookmarkVolumeProperties: volprops,
            kBookmarkCreationOptions: 512,
            kBookmarkWasFileReference: True,
            kBookmarkUserName: "unknown",
            kBookmarkUID: 99,
        }

        if relcount:
            toc[kBookmarkURLLengths] = url_lengths

        return Bookmark([(1, toc)])

    def __repr__(self):
        result = ["Bookmark(["]
        for tid, toc in self.tocs:
            result.append("(0x%x, {\n" % tid)
            for k, v in toc.items():
                if isinstance(k, str):
                    kf = repr(k)
                else:
                    kf = "0x%04x" % k
                result.append(f"  {kf}: {v!r}\n")
            result.append("}),\n")
        result.append("])")

        return "".join(result)



import datetime
import io
import os
import os.path
import struct
import sys
from unicodedata import normalize

if sys.platform == "darwin":
    from . import osx

from .utils import mac_epoch

ALIAS_KIND_FILE = 0
ALIAS_KIND_FOLDER = 1

ALIAS_HFS_VOLUME_SIGNATURE = b"H+"

ALIAS_FILESYSTEM_UDF = "UDF (CD/DVD)"
ALIAS_FILESYSTEM_FAT32 = "FAT32"
ALIAS_FILESYSTEM_EXFAT = "exFAT"
ALIAS_FILESYSTEM_HFSX = "HFSX"
ALIAS_FILESYSTEM_HFSPLUS = "HFS+"
ALIAS_FILESYSTEM_FTP = "FTP"
ALIAS_FILESYSTEM_NTFS = "NTFS"
ALIAS_FILESYSTEM_UNKNOWN = "unknown"

ALIAS_FIXED_DISK = 0
ALIAS_NETWORK_DISK = 1
ALIAS_400KB_FLOPPY_DISK = 2
ALIAS_800KB_FLOPPY_DISK = 3
ALIAS_1_44MB_FLOPPY_DISK = 4
ALIAS_EJECTABLE_DISK = 5

ALIAS_NO_CNID = 0xFFFFFFFF

ALIAS_FSTYPE_MAP = {
    # Version 2 aliases
    b"HX": ALIAS_FILESYSTEM_HFSX,
    b"H+": ALIAS_FILESYSTEM_HFSPLUS,
    # Version 3 aliases
    b"BDcu": ALIAS_FILESYSTEM_UDF,
    b"BDIS": ALIAS_FILESYSTEM_FAT32,
    b"BDxF": ALIAS_FILESYSTEM_EXFAT,
    b"HX\0\0": ALIAS_FILESYSTEM_HFSX,
    b"H+\0\0": ALIAS_FILESYSTEM_HFSPLUS,
    b"KG\0\0": ALIAS_FILESYSTEM_FTP,
    b"NTcu": ALIAS_FILESYSTEM_NTFS,
}


def encode_utf8(s):
    if isinstance(s, bytes):
        return s
    return s.encode("utf-8")


def decode_utf8(s):
    if isinstance(s, bytes):
        return s.decode("utf-8")
    return s


class AppleShareInfo:
    def __init__(self, zone=None, server=None, user=None):
        #: The AppleShare zone
        self.zone = zone
        #: The AFP server
        self.server = server
        #: The username
        self.user = user

    def __repr__(self):
        return "AppleShareInfo({!r},{!r},{!r})".format(
            self.zone, self.server, self.user
        )


class VolumeInfo:
    def __init__(
        self,
        name,
        creation_date,
        fs_type,
        disk_type,
        attribute_flags,
        fs_id,
        appleshare_info=None,
        driver_name=None,
        posix_path=None,
        disk_image_alias=None,
        dialup_info=None,
        network_mount_info=None,
    ):
        #: The name of the volume on which the target resides
        self.name = name

        #: The creation date of the target's volume
        self.creation_date = creation_date

        #: The filesystem type
        #: (for v2 aliases, this is a 2-character code; for v3 aliases, a
        #: 4-character code).
        self.fs_type = fs_type

        #: The type of disk; should be one of
        #:
        #:   * ALIAS_FIXED_DISK
        #:   * ALIAS_NETWORK_DISK
        #:   * ALIAS_400KB_FLOPPY_DISK
        #:   * ALIAS_800KB_FLOPPY_DISK
        #:   * ALIAS_1_44MB_FLOPPY_DISK
        #:   * ALIAS_EJECTABLE_DISK
        self.disk_type = disk_type

        #: Filesystem attribute flags (from HFS volume header)
        self.attribute_flags = attribute_flags

        #: Filesystem identifier
        self.fs_id = fs_id

        #: AppleShare information (for automatic remounting of network shares)
        #: *(optional)*
        self.appleshare_info = appleshare_info

        #: Driver name (*probably* contains a disk driver name on older Macs)
        #: *(optional)*
        self.driver_name = driver_name

        #: POSIX path of the mount point of the target's volume
        #: *(optional)*
        self.posix_path = posix_path

        #: :class:`Alias` object pointing at the disk image on which the
        #: target's volume resides *(optional)*
        self.disk_image_alias = disk_image_alias

        #: Dialup information (for automatic establishment of dialup connections)
        self.dialup_info = dialup_info

        #: Network mount information (for automatic remounting)
        self.network_mount_info = network_mount_info

    @property
    def filesystem_type(self):
        return ALIAS_FSTYPE_MAP.get(self.fs_type, ALIAS_FILESYSTEM_UNKNOWN)

    def __repr__(self):
        args = [
            "name",
            "creation_date",
            "fs_type",
            "disk_type",
            "attribute_flags",
            "fs_id",
        ]
        values = []
        for a in args:
            v = getattr(self, a)
            values.append(repr(v))

        kwargs = [
            "appleshare_info",
            "driver_name",
            "posix_path",
            "disk_image_alias",
            "dialup_info",
            "network_mount_info",
        ]
        for a in kwargs:
            v = getattr(self, a)
            if v is not None:
                values.append(f"{a}={v!r}")
        return "VolumeInfo(%s)" % ",".join(values)


class TargetInfo:
    def __init__(
        self,
        kind,
        filename,
        folder_cnid,
        cnid,
        creation_date,
        creator_code,
        type_code,
        levels_from=-1,
        levels_to=-1,
        folder_name=None,
        cnid_path=None,
        carbon_path=None,
        posix_path=None,
        user_home_prefix_len=None,
    ):
        #: Either ALIAS_KIND_FILE or ALIAS_KIND_FOLDER
        self.kind = kind

        #: The filename of the target
        self.filename = filename

        #: The CNID (Catalog Node ID) of the target's containing folder;
        #: CNIDs are similar to but different than traditional UNIX inode
        #: numbers
        self.folder_cnid = folder_cnid

        #: The CNID (Catalog Node ID) of the target
        self.cnid = cnid

        #: The target's *creation* date.
        self.creation_date = creation_date

        #: The target's Mac creator code (a four-character binary string)
        self.creator_code = creator_code

        #: The target's Mac type code (a four-character binary string)
        self.type_code = type_code

        #: The depth of the alias? Always seems to be -1 on OS X.
        self.levels_from = levels_from

        #: The depth of the target? Always seems to be -1 on OS X.
        self.levels_to = levels_to

        #: The (POSIX) name of the target's containing folder. *(optional)*
        self.folder_name = folder_name

        #: The path from the volume root as a sequence of CNIDs. *(optional)*
        self.cnid_path = cnid_path

        #: The Carbon path of the target *(optional)*
        self.carbon_path = carbon_path

        #: The POSIX path of the target relative to the volume root.  Note
        #: that this may or may not have a leading '/' character, but it is
        #: always relative to the containing volume. *(optional)*
        self.posix_path = posix_path

        #: If the path points into a user's home folder, the number of folders
        #: deep that we go before we get to that home folder. *(optional)*
        self.user_home_prefix_len = user_home_prefix_len

    def __repr__(self):
        args = [
            "kind",
            "filename",
            "folder_cnid",
            "cnid",
            "creation_date",
            "creator_code",
            "type_code",
        ]
        values = []
        for a in args:
            v = getattr(self, a)
            values.append(repr(v))

        if self.levels_from != -1:
            values.append("levels_from=%r" % self.levels_from)
        if self.levels_to != -1:
            values.append("levels_to=%r" % self.levels_to)

        kwargs = [
            "folder_name",
            "cnid_path",
            "carbon_path",
            "posix_path",
            "user_home_prefix_len",
        ]
        for a in kwargs:
            v = getattr(self, a)
            values.append(f"{a}={v!r}")

        return "TargetInfo(%s)" % ",".join(values)


TAG_CARBON_FOLDER_NAME = 0
TAG_CNID_PATH = 1
TAG_CARBON_PATH = 2
TAG_APPLESHARE_ZONE = 3
TAG_APPLESHARE_SERVER_NAME = 4
TAG_APPLESHARE_USERNAME = 5
TAG_DRIVER_NAME = 6
TAG_NETWORK_MOUNT_INFO = 9
TAG_DIALUP_INFO = 10
TAG_UNICODE_FILENAME = 14
TAG_UNICODE_VOLUME_NAME = 15
TAG_HIGH_RES_VOLUME_CREATION_DATE = 16
TAG_HIGH_RES_CREATION_DATE = 17
TAG_POSIX_PATH = 18
TAG_POSIX_PATH_TO_MOUNTPOINT = 19
TAG_RECURSIVE_ALIAS_OF_DISK_IMAGE = 20
TAG_USER_HOME_LENGTH_PREFIX = 21


class Alias:
    def __init__(
        self,
        appinfo=b"\0\0\0\0",
        version=2,
        volume=None,
        target=None,
        extra=[],
    ):
        """Construct a new :class:`Alias` object with the specified
        contents."""

        #: Application specific information (four byte byte-string)
        self.appinfo = appinfo

        #: Version (we support versions 2 and 3)
        self.version = version

        #: A :class:`VolumeInfo` object describing the target's volume
        self.volume = volume

        #: A :class:`TargetInfo` object describing the target
        self.target = target

        #: A list of extra `(tag, value)` pairs
        self.extra = list(extra)

    @classmethod
    def _from_fd(cls, b):
        appinfo, recsize, version = struct.unpack(b">4shh", b.read(8))

        if recsize < 150:
            raise ValueError("Incorrect alias length")

        if version not in (2, 3):
            raise ValueError("Unsupported alias version %u" % version)

        if version == 2:
            (
                kind,  # h
                volname,  # 28p
                voldate,  # I
                fstype,  # 2s
                disktype,  # h
                folder_cnid,  # I
                filename,  # 64p
                cnid,  # I
                crdate,  # I
                creator_code,  # 4s
                type_code,  # 4s
                levels_from,  # h
                levels_to,  # h
                volattrs,  # I
                volfsid,  # 2s
                reserved,  # 10s
            ) = struct.unpack(b">h28pI2shI64pII4s4shhI2s10s", b.read(142))
        else:
            (
                kind,  # h
                voldate_hr,  # Q
                fstype,  # 4s
                disktype,  # h
                folder_cnid,  # I
                cnid,  # I
                crdate_hr,  # Q
                volattrs,  # I
                reserved,  # 14s
            ) = struct.unpack(b">hQ4shIIQI14s", b.read(46))

            volname = b""
            filename = b""
            creator_code = None
            type_code = None
            voldate = voldate_hr / 65536.0
            crdate = crdate_hr / 65536.0

        voldate = mac_epoch + datetime.timedelta(seconds=voldate)
        crdate = mac_epoch + datetime.timedelta(seconds=crdate)

        alias = Alias()
        alias.appinfo = appinfo

        alias.volume = VolumeInfo(
            volname.decode().replace("/", ":"),
            voldate,
            fstype,
            disktype,
            volattrs,
            volfsid,
        )
        alias.target = TargetInfo(
            kind,
            filename.decode().replace("/", ":"),
            folder_cnid,
            cnid,
            crdate,
            creator_code,
            type_code,
        )
        alias.target.levels_from = levels_from
        alias.target.levels_to = levels_to

        tag = struct.unpack(b">h", b.read(2))[0]

        while tag != -1:
            length = struct.unpack(b">h", b.read(2))[0]
            value = b.read(length)
            if length & 1:
                b.read(1)

            if tag == TAG_CARBON_FOLDER_NAME:
                alias.target.folder_name = value.decode().replace("/", ":")
            elif tag == TAG_CNID_PATH:
                alias.target.cnid_path = struct.unpack(">%uI" % (length // 4), value)
            elif tag == TAG_CARBON_PATH:
                alias.target.carbon_path = value
            elif tag == TAG_APPLESHARE_ZONE:
                if alias.volume.appleshare_info is None:
                    alias.volume.appleshare_info = AppleShareInfo()
                alias.volume.appleshare_info.zone = value
            elif tag == TAG_APPLESHARE_SERVER_NAME:
                if alias.volume.appleshare_info is None:
                    alias.volume.appleshare_info = AppleShareInfo()
                alias.volume.appleshare_info.server = value
            elif tag == TAG_APPLESHARE_USERNAME:
                if alias.volume.appleshare_info is None:
                    alias.volume.appleshare_info = AppleShareInfo()
                alias.volume.appleshare_info.user = value
            elif tag == TAG_DRIVER_NAME:
                alias.volume.driver_name = value
            elif tag == TAG_NETWORK_MOUNT_INFO:
                alias.volume.network_mount_info = value
            elif tag == TAG_DIALUP_INFO:
                alias.volume.dialup_info = value
            elif tag == TAG_UNICODE_FILENAME:
                alias.target.filename = value[2:].decode("utf-16be")
            elif tag == TAG_UNICODE_VOLUME_NAME:
                alias.volume.name = value[2:].decode("utf-16be")
            elif tag == TAG_HIGH_RES_VOLUME_CREATION_DATE:
                seconds = struct.unpack(b">Q", value)[0] / 65536.0
                alias.volume.creation_date = mac_epoch + datetime.timedelta(
                    seconds=seconds
                )
            elif tag == TAG_HIGH_RES_CREATION_DATE:
                seconds = struct.unpack(b">Q", value)[0] / 65536.0
                alias.target.creation_date = mac_epoch + datetime.timedelta(
                    seconds=seconds
                )
            elif tag == TAG_POSIX_PATH:
                alias.target.posix_path = value.decode()
            elif tag == TAG_POSIX_PATH_TO_MOUNTPOINT:
                alias.volume.posix_path = value.decode()
            elif tag == TAG_RECURSIVE_ALIAS_OF_DISK_IMAGE:
                alias.volume.disk_image_alias = Alias.from_bytes(value)
            elif tag == TAG_USER_HOME_LENGTH_PREFIX:
                alias.target.user_home_prefix_len = struct.unpack(b">h", value)[0]
            else:
                alias.extra.append((tag, value))

            tag = struct.unpack(b">h", b.read(2))[0]

        return alias

    @classmethod
    def from_bytes(cls, bytes):
        """Construct an :class:`Alias` object given binary Alias data."""
        with io.BytesIO(bytes) as b:
            return cls._from_fd(b)

    @classmethod
    def for_file(cls, path):
        """Create an :class:`Alias` that points at the specified file."""
        if sys.platform != "darwin":
            raise Exception("Not implemented (requires special support)")

        path = encode_utf8(path)

        a = Alias()

        # Find the filesystem
        st = osx.statfs(path)
        vol_path = st.f_mntonname

        # File and folder names in HFS+ are normalized to a form similar to NFD.
        # Must be normalized (NFD->NFC) before use to avoid unicode string comparison issues.
        vol_path = normalize("NFC", vol_path.decode("utf-8")).encode("utf-8")

        # Grab its attributes
        attrs = [osx.ATTR_CMN_CRTIME, osx.ATTR_VOL_NAME, 0, 0, 0]
        volinfo = osx.getattrlist(vol_path, attrs, 0)

        vol_crtime = volinfo[0]
        vol_name = encode_utf8(volinfo[1])

        # Also grab various attributes of the file
        attrs = [
            osx.ATTR_CMN_OBJTYPE
            | osx.ATTR_CMN_CRTIME
            | osx.ATTR_CMN_FNDRINFO
            | osx.ATTR_CMN_FILEID
            | osx.ATTR_CMN_PARENTID,
            0,
            0,
            0,
            0,
        ]
        info = osx.getattrlist(path, attrs, osx.FSOPT_NOFOLLOW)

        if info[0] == osx.VDIR:
            kind = ALIAS_KIND_FOLDER
        else:
            kind = ALIAS_KIND_FILE

        cnid = info[3]
        folder_cnid = info[4]

        dirname, filename = os.path.split(path)

        if dirname == b"" or dirname == b".":
            dirname = os.getcwd()

        foldername = os.path.basename(dirname)

        creation_date = info[1]

        if kind == ALIAS_KIND_FILE:
            creator_code = struct.pack(b"I", info[2].fileInfo.fileCreator)
            type_code = struct.pack(b"I", info[2].fileInfo.fileType)
        else:
            creator_code = b"\0\0\0\0"
            type_code = b"\0\0\0\0"

        a.target = TargetInfo(
            kind, filename, folder_cnid, cnid, creation_date, creator_code, type_code
        )
        a.volume = VolumeInfo(vol_name, vol_crtime, b"H+", ALIAS_FIXED_DISK, 0, b"\0\0")

        a.target.folder_name = foldername
        a.volume.posix_path = vol_path

        rel_path = os.path.relpath(path, vol_path)

        # Leave off the initial '/' if vol_path is '/' (no idea why)
        if vol_path == b"/":
            a.target.posix_path = rel_path
        else:
            a.target.posix_path = b"/" + rel_path

        # Construct the Carbon and CNID paths
        carbon_path = []
        cnid_path = []
        head, tail = os.path.split(rel_path)
        if not tail:
            head, tail = os.path.split(head)
        while head or tail:
            if head:
                attrs = [osx.ATTR_CMN_FILEID, 0, 0, 0, 0]
                info = osx.getattrlist(os.path.join(vol_path, head), attrs, 0)
                cnid_path.append(info[0])
            carbon_tail = tail.replace(b":", b"/")
            carbon_path.insert(0, carbon_tail)
            head, tail = os.path.split(head)

        carbon_path = vol_name + b":" + b":\0".join(carbon_path)

        a.target.carbon_path = carbon_path
        a.target.cnid_path = cnid_path

        return a

    def _to_fd(self, b):
        # We'll come back and fix the length when we're done
        pos = b.tell()
        b.write(struct.pack(b">4shh", self.appinfo, 0, self.version))

        carbon_volname = encode_utf8(self.volume.name).replace(b":", b"/")
        carbon_filename = encode_utf8(self.target.filename).replace(b":", b"/")
        voldate = (self.volume.creation_date - mac_epoch).total_seconds()
        crdate = (self.target.creation_date - mac_epoch).total_seconds()

        if self.version == 2:
            # NOTE: crdate should be in local time, but that's system dependent
            #       (so doing so is ridiculous, and nothing could rely on it).
            b.write(
                struct.pack(
                    b">h28pI2shI64pII4s4shhI2s10s",
                    self.target.kind,  # h
                    carbon_volname,  # 28p
                    int(voldate),  # I
                    self.volume.fs_type,  # 2s
                    self.volume.disk_type,  # h
                    self.target.folder_cnid,  # I
                    carbon_filename,  # 64p
                    self.target.cnid,  # I
                    int(crdate),  # I
                    self.target.creator_code,  # 4s
                    self.target.type_code,  # 4s
                    self.target.levels_from,  # h
                    self.target.levels_to,  # h
                    self.volume.attribute_flags,  # I
                    self.volume.fs_id,  # 2s
                    b"\0" * 10,  # 10s
                )
            )
        else:
            b.write(
                struct.pack(
                    b">hQ4shIIQI14s",
                    self.target.kind,  # h
                    int(voldate * 65536),  # Q
                    self.volume.fs_type,  # 4s
                    self.volume.disk_type,  # h
                    self.target.folder_cnid,  # I
                    self.target.cnid,  # I
                    int(crdate * 65536),  # Q
                    self.volume.attribute_flags,  # I
                    b"\0" * 14,  # 14s
                )
            )

        # Excuse the odd order; we're copying Finder
        if self.target.folder_name:
            carbon_foldername = encode_utf8(self.target.folder_name).replace(b":", b"/")
            b.write(struct.pack(b">hh", TAG_CARBON_FOLDER_NAME, len(carbon_foldername)))
            b.write(carbon_foldername)
            if len(carbon_foldername) & 1:
                b.write(b"\0")

        b.write(
            struct.pack(
                b">hhQhhQ",
                TAG_HIGH_RES_VOLUME_CREATION_DATE,
                8,
                int(voldate * 65536),
                TAG_HIGH_RES_CREATION_DATE,
                8,
                int(crdate * 65536),
            )
        )

        if self.target.cnid_path:
            cnid_path = struct.pack(
                ">%uI" % len(self.target.cnid_path), *self.target.cnid_path
            )
            b.write(struct.pack(b">hh", TAG_CNID_PATH, len(cnid_path)))
            b.write(cnid_path)

        if self.target.carbon_path:
            carbon_path = encode_utf8(self.target.carbon_path)
            b.write(struct.pack(b">hh", TAG_CARBON_PATH, len(carbon_path)))
            b.write(carbon_path)
            if len(carbon_path) & 1:
                b.write(b"\0")

        if self.volume.appleshare_info:
            ai = self.volume.appleshare_info
            if ai.zone:
                b.write(struct.pack(b">hh", TAG_APPLESHARE_ZONE, len(ai.zone)))
                b.write(ai.zone)
                if len(ai.zone) & 1:
                    b.write(b"\0")
            if ai.server:
                b.write(struct.pack(b">hh", TAG_APPLESHARE_SERVER_NAME, len(ai.server)))
                b.write(ai.server)
                if len(ai.server) & 1:
                    b.write(b"\0")
            if ai.username:
                b.write(struct.pack(b">hh", TAG_APPLESHARE_USERNAME, len(ai.username)))
                b.write(ai.username)
                if len(ai.username) & 1:
                    b.write(b"\0")

        if self.volume.driver_name:
            driver_name = encode_utf8(self.volume.driver_name)
            b.write(struct.pack(b">hh", TAG_DRIVER_NAME, len(driver_name)))
            b.write(driver_name)
            if len(driver_name) & 1:
                b.write(b"\0")

        if self.volume.network_mount_info:
            b.write(
                struct.pack(
                    b">hh", TAG_NETWORK_MOUNT_INFO, len(self.volume.network_mount_info)
                )
            )
            b.write(self.volume.network_mount_info)
            if len(self.volume.network_mount_info) & 1:
                b.write(b"\0")

        if self.volume.dialup_info:
            b.write(
                struct.pack(
                    b">hh", TAG_DIALUP_INFO, len(self.volume.network_mount_info)
                )
            )
            b.write(self.volume.network_mount_info)
            if len(self.volume.network_mount_info) & 1:
                b.write(b"\0")

        utf16 = decode_utf8(self.target.filename).replace(":", "/").encode("utf-16-be")
        b.write(
            struct.pack(b">hhh", TAG_UNICODE_FILENAME, len(utf16) + 2, len(utf16) // 2)
        )
        b.write(utf16)

        utf16 = decode_utf8(self.volume.name).replace(":", "/").encode("utf-16-be")
        b.write(
            struct.pack(
                b">hhh", TAG_UNICODE_VOLUME_NAME, len(utf16) + 2, len(utf16) // 2
            )
        )
        b.write(utf16)

        if self.target.posix_path:
            posix_path = encode_utf8(self.target.posix_path)
            b.write(struct.pack(b">hh", TAG_POSIX_PATH, len(posix_path)))
            b.write(posix_path)
            if len(posix_path) & 1:
                b.write(b"\0")

        if self.volume.posix_path:
            posix_path = encode_utf8(self.volume.posix_path)
            b.write(struct.pack(b">hh", TAG_POSIX_PATH_TO_MOUNTPOINT, len(posix_path)))
            b.write(posix_path)
            if len(posix_path) & 1:
                b.write(b"\0")

        if self.volume.disk_image_alias:
            d = self.volume.disk_image_alias.to_bytes()
            b.write(struct.pack(b">hh", TAG_RECURSIVE_ALIAS_OF_DISK_IMAGE, len(d)))
            b.write(d)
            if len(d) & 1:
                b.write(b"\0")

        if self.target.user_home_prefix_len is not None:
            b.write(
                struct.pack(
                    b">hhh",
                    TAG_USER_HOME_LENGTH_PREFIX,
                    2,
                    self.target.user_home_prefix_len,
                )
            )

        for t, v in self.extra:
            b.write(struct.pack(b">hh", t, len(v)))
            b.write(v)
            if len(v) & 1:
                b.write(b"\0")

        b.write(struct.pack(b">hh", -1, 0))

        blen = b.tell() - pos
        b.seek(pos + 4, os.SEEK_SET)
        b.write(struct.pack(b">h", blen))

    def to_bytes(self):
        """Returns the binary representation for this :class:`Alias`."""
        with io.BytesIO() as b:
            self._to_fd(b)
            return b.getvalue()

    def __str__(self):
        return "<Alias target=%s>" % self.target.filename

    def __repr__(self):
        values = []
        if self.appinfo != b"\0\0\0\0":
            values.append("appinfo=%r" % self.appinfo)
        if self.version != 2:
            values.append("version=%r" % self.version)
        if self.volume is not None:
            values.append("volume=%r" % self.volume)
        if self.target is not None:
            values.append("target=%r" % self.target)
        if self.extra:
            values.append("extra=%r" % self.extra)
        return "Alias(%s)" % ",".join(values)
